#!/usr/bin/env python

"""
<Program Name>
  interface.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  January 5, 2017.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provide an interface to the cryptography functions available in
  securesystemslib.  The interface can be used with the Python interpreter in
  interactive mode, or imported directly into a Python module.  See
  'securesystemslib/README' for the complete guide to using 'interface.py'.
"""

import getpass
import json
import logging
import os
import sys
import tempfile

from securesystemslib import (
    KEY_TYPE_ECDSA,
    KEY_TYPE_ED25519,
    KEY_TYPE_RSA,
    exceptions,
    formats,
    keys,
    settings,
    util,
)
from securesystemslib.storage import FilesystemBackend

logger = logging.getLogger(__name__)

# Recommended RSA key sizes:
# https://en.wikipedia.org/wiki/Key_size#Asymmetric_algorithm_key_lengths
# Based on the above, RSA keys of size 3072 bits are expected to provide
# security through 2031 and beyond.
DEFAULT_RSA_KEY_BITS = 3072


def get_password(prompt="Password: ", confirm=False):
    """Prompts user to enter a password.

    Arguments:
      prompt (optional): A text displayed on the prompt (stderr).
      confirm (optional): A boolean indicating if the user needs to enter the
          same password twice.

    Returns:
      The password entered on the prompt.

    """
    formats.TEXT_SCHEMA.check_match(prompt)
    formats.BOOLEAN_SCHEMA.check_match(confirm)

    while True:
        # getpass() prompts the user for a password without echoing
        # the user input.
        password = getpass.getpass(prompt, sys.stderr)

        if not confirm:
            return password
        password2 = getpass.getpass("Confirm: ", sys.stderr)

        if password == password2:  # pylint: disable=no-else-return
            return password

        else:
            print("Mismatch; try again.")


def _get_key_file_encryption_password(password, prompt, path):
    """Encryption password helper for `_generate_and_write_*_keypair` functions.

    Combinations of 'password' and 'prompt' -> result (explanation)
    ----------------------------------------------------------------
    None                  False -> return None (clear non-encryption desire)
    "<non-empty string>"  False -> return password (clear encryption desire)
    <anything else>       False -> raise (bad pw type, unclear encryption desire)
    <not None>            True  -> raise (unclear password/prompt precedence)
    None                  True  -> prompt and return password if entered and None
                                   otherwise (users on the prompt can only
                                   indicate desire to not encrypt by entering no
                                   password)
    """
    formats.BOOLEAN_SCHEMA.check_match(prompt)

    # We don't want to decide which takes precedence so we fail
    if password is not None and prompt:
        raise ValueError("passing 'password' and 'prompt=True' is not allowed")

    # Prompt user for password and confirmation
    if prompt:
        password = get_password(
            "enter password to encrypt private key file "
            "'" + str(path) + "' (leave empty if key "
            "should not be encrypted): ",
            confirm=True,
        )

        # Treat empty password as no password. A user on the prompt can only
        # indicate the desire to not encrypt by entering no password.
        if not len(password):  # pylint: disable=use-implicit-booleaness-not-len
            return None

    if password is not None:
        formats.PASSWORD_SCHEMA.check_match(password)

        # Fail on empty passed password. A caller should pass None to indicate the
        # desire to not encrypt.
        if not len(password):  # pylint: disable=use-implicit-booleaness-not-len
            raise ValueError(
                "encryption password must be 1 or more characters long"
            )

    return password


def _get_key_file_decryption_password(password, prompt, path):
    """Decryption password helper for `import_*_privatekey_from_file` functions.

    Combinations of 'password' and 'prompt' -> result (explanation)
    ----------------------------------------------------------------
    None             False -> return None (clear non-decryption desire)
    "<any string>"   False -> return password (clear decryption desire)
    <anything else>  False -> raise (bad pw type, unclear decryption desire)
    <not None>       True  -> raise (unclear password/prompt precedence)
    None             True  -> prompt and return password if entered and None
                              otherwise (users on the prompt can only indicate
                              desire to not decrypt by entering no password)

    """
    formats.BOOLEAN_SCHEMA.check_match(prompt)

    # We don't want to decide which takes precedence so we fail
    if password is not None and prompt:
        raise ValueError("passing 'password' and 'prompt=True' is not allowed")

    # Prompt user for password
    if prompt:
        password = get_password(
            "enter password to decrypt private key file "
            "'" + str(path) + "' "
            "(leave empty if key not encrypted): ",
            confirm=False,
        )

        # Treat empty password as no password. A user on the prompt can only
        # indicate the desire to not decrypt by entering no password.
        if not len(password):  # pylint: disable=use-implicit-booleaness-not-len
            return None

    if password is not None:
        formats.PASSWORD_SCHEMA.check_match(password)
        # No additional vetting needed. Decryption will show if it was correct.

    return password


def _generate_and_write_rsa_keypair(
    filepath=None, bits=DEFAULT_RSA_KEY_BITS, password=None, prompt=False
):
    """Generates RSA key pair and writes PEM-encoded keys to disk.

    If a password is passed or entered on the prompt, the private key is
    encrypted. According to the documentation of the used pyca/cryptography
    library, encryption is performed "using the best available encryption for a
    given key's backend", which "is a curated encryption choice and the algorithm
    may change over time."  The private key is written in PKCS#1 and the public
    key in X.509 SubjectPublicKeyInfo format.

    NOTE: A signing scheme can be assigned on key import (see import functions).

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.
      bits (optional): The number of bits of the generated RSA key.
      password (optional): An encryption password.
      prompt (optional): A boolean indicating if the user should be prompted
          for an encryption password. If the user enters an empty password, the
          key is not encrypted.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: An empty string is passed as 'password', or both a 'password'
          is passed and 'prompt' is true.
      StorageError: Key files cannot be written.

    Side Effects:
      Prompts user for a password if 'prompt' is True.
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    formats.RSAKEYBITS_SCHEMA.check_match(bits)

    # Generate private RSA key and extract public and private both in PEM
    rsa_key = keys.generate_rsa_key(bits)
    public = rsa_key["keyval"]["public"]
    private = rsa_key["keyval"]["private"]

    # Use passed 'filepath' or keyid as file name
    if not filepath:
        filepath = os.path.join(os.getcwd(), rsa_key["keyid"])

    formats.PATH_SCHEMA.check_match(filepath)

    password = _get_key_file_encryption_password(password, prompt, filepath)

    # Encrypt the private key if a 'password' was passed or entered on the prompt
    if password is not None:
        private = keys.create_rsa_encrypted_pem(private, password)

    # Create intermediate directories as required
    util.ensure_parent_dir(filepath)

    # Write PEM-encoded public key to <filepath>.pub
    file_object = tempfile.TemporaryFile()
    file_object.write(public.encode("utf-8"))
    util.persist_temp_file(file_object, filepath + ".pub")

    # Write PEM-encoded private key to <filepath>
    file_object = tempfile.TemporaryFile()
    file_object.write(private.encode("utf-8"))
    util.persist_temp_file(file_object, filepath, restrict=True)

    return filepath


def generate_and_write_rsa_keypair(
    password, filepath=None, bits=DEFAULT_RSA_KEY_BITS
):
    """Generates RSA key pair and writes PEM-encoded keys to disk.

    The private key is encrypted using the best available encryption algorithm
    chosen by 'pyca/cryptography', which may change over time. The private key is
    written in PKCS#1 and the public key in X.509 SubjectPublicKeyInfo format.

    NOTE: A signing scheme can be assigned on key import (see import functions).

    Arguments:
      password: An encryption password.
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.
      bits (optional): The number of bits of the generated RSA key.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: An empty string is passed as 'password'.
      StorageError: Key files cannot be written.

    Side Effects:
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    formats.PASSWORD_SCHEMA.check_match(password)
    return _generate_and_write_rsa_keypair(
        filepath=filepath, bits=bits, password=password, prompt=False
    )


def generate_and_write_rsa_keypair_with_prompt(
    filepath=None, bits=DEFAULT_RSA_KEY_BITS
):
    """Generates RSA key pair and writes PEM-encoded keys to disk.

    The private key is encrypted with a password entered on the prompt, using the
    best available encryption algorithm chosen by 'pyca/cryptography', which may
    change over time. The private key is written in PKCS#1 and the public key in
    X.509 SubjectPublicKeyInfo format.

    NOTE: A signing scheme can be assigned on key import (see import functions).

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.
      bits (optional): The number of bits of the generated RSA key.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key files cannot be written.

    Side Effects:
      Prompts user for a password.
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    return _generate_and_write_rsa_keypair(
        filepath=filepath, bits=bits, password=None, prompt=True
    )


def generate_and_write_unencrypted_rsa_keypair(
    filepath=None, bits=DEFAULT_RSA_KEY_BITS
):
    """Generates RSA key pair and writes PEM-encoded keys to disk.

    The private key is written in PKCS#1 and the public key in X.509
    SubjectPublicKeyInfo format.

    NOTE: A signing scheme can be assigned on key import (see import functions).

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.
      bits (optional): The number of bits of the generated RSA key.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key files cannot be written.

    Side Effects:
      Writes unencrypted key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    return _generate_and_write_rsa_keypair(
        filepath=filepath, bits=bits, password=None, prompt=False
    )


def import_rsa_privatekey_from_file(
    filepath,
    password=None,
    scheme="rsassa-pss-sha256",
    prompt=False,
    storage_backend=None,
):
    """Imports PEM-encoded RSA private key from file storage.

    The expected key format is PKCS#1. If a password is passed or entered on the
    prompt, the private key is decrypted, otherwise it is treated as unencrypted.

    Arguments:
      filepath: The path to read the file from.
      password (optional): A password to decrypt the key.
      scheme (optional): The signing scheme assigned to the returned key object.
          See RSA_SCHEME_SCHEMA for available signing schemes.
      prompt (optional): A boolean indicating if the user should be prompted
          for a decryption password. If the user enters an empty password, the
          key is not decrypted.
      storage_backend (optional): An object implementing StorageBackendInterface.
          If not passed a default FilesystemBackend will be used.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: Both a 'password' is passed and 'prompt' is true.
      StorageError: Key file cannot be read.
      CryptoError: Key cannot be parsed.

    Returns:
      An RSA private key object conformant with 'RSAKEY_SCHEMA'.

    """
    formats.PATH_SCHEMA.check_match(filepath)
    formats.RSA_SCHEME_SCHEMA.check_match(scheme)

    password = _get_key_file_decryption_password(password, prompt, filepath)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    with storage_backend.get(filepath) as file_object:
        pem_key = file_object.read().decode("utf-8")

    # Optionally decrypt and convert PEM-encoded key to 'RSAKEY_SCHEMA' format
    rsa_key = keys.import_rsakey_from_private_pem(pem_key, scheme, password)

    return rsa_key


def import_rsa_publickey_from_file(
    filepath, scheme="rsassa-pss-sha256", storage_backend=None
):
    """Imports PEM-encoded RSA public key from file storage.

    The expected key format is X.509 SubjectPublicKeyInfo.

    Arguments:
      filepath: The path to read the file from.
      scheme (optional): The signing scheme assigned to the returned key object.
          See RSA_SCHEME_SCHEMA for available signing schemes.
      storage_backend (optional): An object implementing StorageBackendInterface.
          If not passed a default FilesystemBackend will be used.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key file cannot be read.
      Error: Public key is malformed.

    Returns:
      An RSA public key object conformant with 'RSAKEY_SCHEMA'.

    """
    formats.PATH_SCHEMA.check_match(filepath)
    formats.RSA_SCHEME_SCHEMA.check_match(scheme)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    with storage_backend.get(filepath) as file_object:
        rsa_pubkey_pem = file_object.read().decode("utf-8")

    # Convert PEM-encoded key to 'RSAKEY_SCHEMA' format
    try:
        rsakey_dict = keys.import_rsakey_from_public_pem(rsa_pubkey_pem, scheme)

    except exceptions.FormatError as e:
        raise exceptions.Error(
            "Cannot import improperly formatted" " PEM file." + repr(str(e))
        )

    return rsakey_dict


def _generate_and_write_ed25519_keypair(
    filepath=None, password=None, prompt=False
):
    """Generates ed25519 key pair and writes custom JSON-formatted keys to disk.

    If a password is passed or entered on the prompt, the private key is
    encrypted using AES-256 in CTR mode, with the password strengthened in
    PBKDF2-HMAC-SHA256.

    NOTE: The custom key format includes 'ed25519' as signing scheme.

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.
      password (optional): An encryption password.
      prompt (optional): A boolean indicating if the user should be prompted
          for an encryption password. If the user enters an empty password, the
          key is not encrypted.

    Raises:
      UnsupportedLibraryError: pyca/pynacl or pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: An empty string is passed as 'password', or both a 'password'
          is passed and 'prompt' is true.
      StorageError: Key files cannot be written.

    Side Effects:
      Prompts user for a password if 'prompt' is True.
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    ed25519_key = keys.generate_ed25519_key()

    # Use passed 'filepath' or keyid as file name
    if not filepath:
        filepath = os.path.join(os.getcwd(), ed25519_key["keyid"])

    formats.PATH_SCHEMA.check_match(filepath)

    password = _get_key_file_encryption_password(password, prompt, filepath)

    # Create intermediate directories as required
    util.ensure_parent_dir(filepath)

    # Use custom JSON format for ed25519 keys on-disk
    keytype = ed25519_key["keytype"]
    keyval = ed25519_key["keyval"]
    scheme = ed25519_key["scheme"]
    ed25519key_metadata_format = keys.format_keyval_to_metadata(
        keytype, scheme, keyval, private=False
    )

    # Write public key to <filepath>.pub
    file_object = tempfile.TemporaryFile()
    file_object.write(json.dumps(ed25519key_metadata_format).encode("utf-8"))
    util.persist_temp_file(file_object, filepath + ".pub")

    # Encrypt private key if we have a password, store as JSON string otherwise
    if password is not None:
        ed25519_key = keys.encrypt_key(ed25519_key, password)
    else:
        ed25519_key = json.dumps(ed25519_key)

    # Write private key to <filepath>
    file_object = tempfile.TemporaryFile()
    file_object.write(ed25519_key.encode("utf-8"))
    util.persist_temp_file(file_object, filepath, restrict=True)

    return filepath


def generate_and_write_ed25519_keypair(password, filepath=None):
    """Generates ed25519 key pair and writes custom JSON-formatted keys to disk.

    The private key is encrypted using AES-256 in CTR mode, with the passed
    password strengthened in PBKDF2-HMAC-SHA256.

    NOTE: The custom key format includes 'ed25519' as signing scheme.

    Arguments:
      password: An encryption password.
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.

    Raises:
      UnsupportedLibraryError: pyca/pynacl or pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: An empty string is passed as 'password'.
      StorageError: Key files cannot be written.

    Side Effects:
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    formats.PASSWORD_SCHEMA.check_match(password)
    return _generate_and_write_ed25519_keypair(
        filepath=filepath, password=password, prompt=False
    )


def generate_and_write_ed25519_keypair_with_prompt(filepath=None):
    """Generates ed25519 key pair and writes custom JSON-formatted keys to disk.

    The private key is encrypted using AES-256 in CTR mode, with the password
    entered on the prompt strengthened in PBKDF2-HMAC-SHA256.

    NOTE: The custom key format includes 'ed25519' as signing scheme.

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.

    Raises:
      UnsupportedLibraryError: pyca/pynacl or pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key files cannot be written.

    Side Effects:
      Prompts user for a password.
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    return _generate_and_write_ed25519_keypair(
        filepath=filepath, password=None, prompt=True
    )


def generate_and_write_unencrypted_ed25519_keypair(filepath=None):
    """Generates ed25519 key pair and writes custom JSON-formatted keys to disk.

    NOTE: The custom key format includes 'ed25519' as signing scheme.

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.

    Raises:
      UnsupportedLibraryError: pyca/pynacl or pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key files cannot be written.

    Side Effects:
      Writes unencrypted key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    return _generate_and_write_ed25519_keypair(
        filepath=filepath, password=None, prompt=False
    )


def import_ed25519_publickey_from_file(filepath):
    """Imports custom JSON-formatted ed25519 public key from disk.

    NOTE: The signing scheme is set at key generation (see generate function).

    Arguments:
      filepath: The path to read the file from.

    Raises:
      FormatError: Argument is malformed.
      StorageError: Key file cannot be read.
      Error: Public key is malformed.

    Returns:
      An ed25519 public key object conformant with 'ED25519KEY_SCHEMA'.

    """
    formats.PATH_SCHEMA.check_match(filepath)

    # Load custom on-disk JSON formatted key and convert to its custom in-memory
    # dict key representation
    ed25519_key_metadata = util.load_json_file(filepath)
    ed25519_key, _ = keys.format_metadata_to_key(ed25519_key_metadata)

    # Check that the generic loading functions indeed loaded an ed25519 key
    if ed25519_key["keytype"] != "ed25519":
        message = "Invalid key type loaded: " + repr(ed25519_key["keytype"])
        raise exceptions.FormatError(message)

    return ed25519_key


def import_ed25519_privatekey_from_file(
    filepath, password=None, prompt=False, storage_backend=None
):
    """Imports custom JSON-formatted ed25519 private key from file storage.

    If a password is passed or entered on the prompt, the private key is
    decrypted, otherwise it is treated as unencrypted.

    NOTE: The signing scheme is set at key generation (see generate function).

    Arguments:
      filepath: The path to read the file from.
      password (optional): A password to decrypt the key.
      prompt (optional): A boolean indicating if the user should be prompted
          for a decryption password. If the user enters an empty password, the
          key is not decrypted.
      storage_backend (optional): An object implementing StorageBackendInterface.
          If not passed a default FilesystemBackend will be used.


    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: Both a 'password' is passed and 'prompt' is true.
      StorageError: Key file cannot be read.
      Error, CryptoError: Key cannot be parsed.


    Returns:
      An ed25519 private key object conformant with 'ED25519KEY_SCHEMA'.

    """
    formats.PATH_SCHEMA.check_match(filepath)
    password = _get_key_file_decryption_password(password, prompt, filepath)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    with storage_backend.get(filepath) as file_object:
        json_str = file_object.read()

        # Load custom on-disk JSON formatted key and convert to its custom
        # in-memory dict key representation, decrypting it if password is not None
        return keys.import_ed25519key_from_private_json(
            json_str, password=password
        )


def _generate_and_write_ecdsa_keypair(
    filepath=None, password=None, prompt=False
):
    """Generates ecdsa key pair and writes custom JSON-formatted keys to disk.

    If a password is passed or entered on the prompt, the private key is
    encrypted using AES-256 in CTR mode, with the password strengthened in
    PBKDF2-HMAC-SHA256.

    NOTE: The custom key format includes 'ecdsa-sha2-nistp256' as signing scheme.

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.
      password (optional): An encryption password.
      prompt (optional): A boolean indicating if the user should be prompted
          for an encryption password. If the user enters an empty password, the
          key is not encrypted.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: An empty string is passed as 'password', or both a 'password'
          is passed and 'prompt' is true.
      StorageError: Key files cannot be written.

    Side Effects:
      Prompts user for a password if 'prompt' is True.
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    ecdsa_key = keys.generate_ecdsa_key()

    # Use passed 'filepath' or keyid as file name
    if not filepath:
        filepath = os.path.join(os.getcwd(), ecdsa_key["keyid"])

    formats.PATH_SCHEMA.check_match(filepath)

    password = _get_key_file_encryption_password(password, prompt, filepath)

    # Create intermediate directories as required
    util.ensure_parent_dir(filepath)

    # Use custom JSON format for ecdsa keys on-disk
    keytype = ecdsa_key["keytype"]
    keyval = ecdsa_key["keyval"]
    scheme = ecdsa_key["scheme"]
    ecdsakey_metadata_format = keys.format_keyval_to_metadata(
        keytype, scheme, keyval, private=False
    )

    # Write public key to <filepath>.pub
    file_object = tempfile.TemporaryFile()
    file_object.write(json.dumps(ecdsakey_metadata_format).encode("utf-8"))
    util.persist_temp_file(file_object, filepath + ".pub")

    # Encrypt private key if we have a password, store as JSON string otherwise
    if password is not None:
        ecdsa_key = keys.encrypt_key(ecdsa_key, password)
    else:
        ecdsa_key = json.dumps(ecdsa_key)

    # Write private key to <filepath>
    file_object = tempfile.TemporaryFile()
    file_object.write(ecdsa_key.encode("utf-8"))
    util.persist_temp_file(file_object, filepath, restrict=True)

    return filepath


def generate_and_write_ecdsa_keypair(password, filepath=None):
    """Generates ecdsa key pair and writes custom JSON-formatted keys to disk.

    The private key is encrypted using AES-256 in CTR mode, with the passed
    password strengthened in PBKDF2-HMAC-SHA256.

    NOTE: The custom key format includes 'ecdsa-sha2-nistp256' as signing scheme.

    Arguments:
      password: An encryption password.
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: An empty string is passed as 'password'.
      StorageError: Key files cannot be written.

    Side Effects:
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    formats.PASSWORD_SCHEMA.check_match(password)
    return _generate_and_write_ecdsa_keypair(
        filepath=filepath, password=password, prompt=False
    )


def generate_and_write_ecdsa_keypair_with_prompt(filepath=None):
    """Generates ecdsa key pair and writes custom JSON-formatted keys to disk.

    The private key is encrypted using AES-256 in CTR mode, with the password
    entered on the prompt strengthened in PBKDF2-HMAC-SHA256.

    NOTE: The custom key format includes 'ecdsa-sha2-nistp256' as signing scheme.

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key files cannot be written.

    Side Effects:
      Prompts user for a password.
      Writes key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    return _generate_and_write_ecdsa_keypair(
        filepath=filepath, password=None, prompt=True
    )


def generate_and_write_unencrypted_ecdsa_keypair(filepath=None):
    """Generates ecdsa key pair and writes custom JSON-formatted keys to disk.

    NOTE: The custom key format includes 'ecdsa-sha2-nistp256' as signing scheme.

    Arguments:
      filepath (optional): The path to write the private key to. If not passed,
          the key is written to CWD using the keyid as filename. The public key
          is written to the same path as the private key using the suffix '.pub'.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      StorageError: Key files cannot be written.

    Side Effects:
      Writes unencrypted key files to disk.
      Overwrites files if they already exist.

    Returns:
      The private key filepath.

    """
    return _generate_and_write_ecdsa_keypair(
        filepath=filepath, password=None, prompt=False
    )


def import_ecdsa_publickey_from_file(filepath):
    """Imports custom JSON-formatted ecdsa public key from disk.

    NOTE: The signing scheme is set at key generation (see generate function).

    Arguments:
      filepath: The path to read the file from.

    Raises:
      FormatError: Argument is malformed.
      StorageError: Key file cannot be read.
      Error: Public key is malformed.

    Returns:
      An ecdsa public key object conformant with 'ECDSAKEY_SCHEMA'.

    """
    formats.PATH_SCHEMA.check_match(filepath)

    # Load custom on-disk JSON formatted key and convert to its custom in-memory
    # dict key representation
    ecdsa_key_metadata = util.load_json_file(filepath)
    ecdsa_key, _ = keys.format_metadata_to_key(ecdsa_key_metadata)

    return ecdsa_key


def import_ecdsa_privatekey_from_file(
    filepath, password=None, prompt=False, storage_backend=None
):
    """Imports custom JSON-formatted ecdsa private key from file storage.

    If a password is passed or entered on the prompt, the private key is
    decrypted, otherwise it is treated as unencrypted.

    NOTE: The signing scheme is set at key generation (see generate function).

    Arguments:
      filepath: The path to read the file from.
      password (optional): A password to decrypt the key.
      prompt (optional): A boolean indicating if the user should be prompted
          for a decryption password. If the user enters an empty password, the
          key is not decrypted.
      storage_backend (optional): An object implementing StorageBackendInterface.
          If not passed a default FilesystemBackend will be used.

    Raises:
      UnsupportedLibraryError: pyca/cryptography is not available.
      FormatError: Arguments are malformed.
      ValueError: Both a 'password' is passed and 'prompt' is true.
      StorageError: Key file cannot be read.
      Error, CryptoError: Key cannot be parsed.

    Returns:
      An ecdsa private key object conformant with 'ED25519KEY_SCHEMA'.

    """
    formats.PATH_SCHEMA.check_match(filepath)

    password = _get_key_file_decryption_password(password, prompt, filepath)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    with storage_backend.get(filepath) as file_object:
        key_data = file_object.read().decode("utf-8")

    # Decrypt private key if we have a password, directly load JSON otherwise
    if password is not None:
        key_object = keys.decrypt_key(key_data, password)
    else:
        key_object = util.load_json_string(key_data)

    # Raise an exception if an unexpected key type is imported.
    # NOTE: we support keytype's of ecdsa-sha2-nistp256 and ecdsa-sha2-nistp384
    # in order to support key files generated with older versions of
    # securesystemslib. At some point this backwards compatibility should be
    # removed.
    if key_object["keytype"] not in [
        "ecdsa",
        "ecdsa-sha2-nistp256",
        "ecdsa-sha2-nistp384",
    ]:
        message = "Invalid key type loaded: " + repr(key_object["keytype"])
        raise exceptions.FormatError(message)

    # Add "keyid_hash_algorithms" so that equal ecdsa keys with different keyids
    # can be associated using supported keyid_hash_algorithms.
    key_object["keyid_hash_algorithms"] = settings.HASH_ALGORITHMS

    return key_object


def import_publickeys_from_file(filepaths, key_types=None):
    """Imports multiple public keys from files.

    NOTE: The default signing scheme 'rsassa-pss-sha256' is assigned to RSA keys.
    Use 'import_rsa_publickey_from_file' to specify any other than the default
    signing scheme for an RSA key. ed25519 and ecdsa keys have the signing scheme
    included in the custom key format (see generate functions).

    Arguments:
      filepaths: A list of paths to public key files.
      key_types (optional): A list of types of keys to be imported associated
          with filepaths by index. Must be one of KEY_TYPE_RSA, KEY_TYPE_ED25519
          or KEY_TYPE_ECDSA. If not specified, all keys are assumed to be
          KEY_TYPE_RSA.

    Raises:
      TypeError: filepaths or 'key_types' (if passed) is not iterable.
      FormatError: Argument are malformed, or 'key_types' is passed and does not
          have the same length as 'filepaths' or contains an unsupported type.
      UnsupportedLibraryError: pyca/cryptography is not available.
      StorageError: Key file cannot be read.
      Error: Public key is malformed.

    Returns:
      A dict of public keys in KEYDICT_SCHEMA format.

    """
    if key_types is None:
        key_types = [KEY_TYPE_RSA] * len(filepaths)

    if len(key_types) != len(filepaths):
        raise exceptions.FormatError(
            "Pass equal amount of 'filepaths' (got {}) and 'key_types (got {}), "  # pylint: disable=consider-using-f-string
            "or no 'key_types' at all to default to '{}'.".format(
                len(filepaths), len(key_types), KEY_TYPE_RSA
            )
        )

    key_dict = {}
    for idx, filepath in enumerate(filepaths):
        if key_types[idx] == KEY_TYPE_ED25519:
            key = import_ed25519_publickey_from_file(filepath)

        elif key_types[idx] == KEY_TYPE_RSA:
            key = import_rsa_publickey_from_file(filepath)

        elif key_types[idx] == KEY_TYPE_ECDSA:
            key = import_ecdsa_publickey_from_file(filepath)

        else:
            raise exceptions.FormatError(
                "Unsupported key type '{}'. Must be '{}', '{}' or '{}'.".format(  # pylint: disable=consider-using-f-string
                    key_types[idx],
                    KEY_TYPE_RSA,
                    KEY_TYPE_ED25519,
                    KEY_TYPE_ECDSA,
                )
            )

        key_dict[key["keyid"]] = key

    return key_dict


def import_privatekey_from_file(
    filepath, key_type=None, password=None, prompt=False
):
    """Imports private key from file.

    If a password is passed or entered on the prompt, the private key is
    decrypted, otherwise it is treated as unencrypted.

    NOTE: The default signing scheme 'rsassa-pss-sha256' is assigned to RSA keys.
    Use 'import_rsa_privatekey_from_file' to specify any other than the default
    signing scheme for an RSA key. ed25519 and ecdsa keys have the signing scheme
    included in the custom key format (see generate functions).

    Arguments:
      filepath: The path to read the file from.
      key_type (optional): One of KEY_TYPE_RSA, KEY_TYPE_ED25519 or
          KEY_TYPE_ECDSA. Default is KEY_TYPE_RSA.
      password (optional): A password to decrypt the key.
      prompt (optional): A boolean indicating if the user should be prompted
          for a decryption password. If the user enters an empty password, the
          key is not decrypted.

    Raises:
      FormatError: Arguments are malformed or 'key_type' is not supported.
      ValueError: Both a 'password' is passed and 'prompt' is true.
      UnsupportedLibraryError: pyca/cryptography is not available.
      StorageError: Key file cannot be read.
      Error, CryptoError: Key cannot be parsed.

    Returns:
      A private key object conformant with one of 'ED25519KEY_SCHEMA',
      'RSAKEY_SCHEMA' or 'ECDSAKEY_SCHEMA'.

    """
    if key_type is None:
        key_type = KEY_TYPE_RSA

    if key_type == KEY_TYPE_ED25519:  # pylint: disable=no-else-return
        return import_ed25519_privatekey_from_file(
            filepath, password=password, prompt=prompt
        )

    elif key_type == KEY_TYPE_RSA:
        return import_rsa_privatekey_from_file(
            filepath, password=password, prompt=prompt
        )

    elif key_type == KEY_TYPE_ECDSA:
        return import_ecdsa_privatekey_from_file(
            filepath, password=password, prompt=prompt
        )

    else:
        raise exceptions.FormatError(
            "Unsupported key type '{}'. Must be '{}', '{}' or '{}'.".format(  # pylint: disable=consider-using-f-string
                key_type, KEY_TYPE_RSA, KEY_TYPE_ED25519, KEY_TYPE_ECDSA
            )
        )


if __name__ == "__main__":
    # The interactive sessions of the documentation strings can
    # be tested by running interface.py as a standalone module:
    # $ python interface.py.
    import doctest

    doctest.testmod()



"""Hardware Security Module (HSM) Signer

Uses PKCS#11/Cryptoki API to create signatures with HSMs (e.g. YubiKey) and to export
the related public keys.

"""
import binascii
from contextlib import contextmanager
from typing import Optional, Tuple
from urllib import parse

from securesystemslib import KEY_TYPE_ECDSA
from securesystemslib.exceptions import UnsupportedLibraryError
from securesystemslib.keys import _get_keyid
from securesystemslib.signer._key import Key, SSlibKey
from securesystemslib.signer._signature import Signature
from securesystemslib.signer._signer import SecretsHandler, Signer

# pylint: disable=wrong-import-position
CRYPTO_IMPORT_ERROR = None
try:
    from cryptography.hazmat.primitives import serialization
    from cryptography.hazmat.primitives.asymmetric.ec import (
        SECP256R1,
        SECP384R1,
        EllipticCurvePublicKey,
        ObjectIdentifier,
        get_curve_for_oid,
    )
    from cryptography.hazmat.primitives.asymmetric.utils import (
        encode_dss_signature,
    )

    # TODO: Don't hardcode schemes
    _SCHEME_FOR_CURVE = {
        SECP256R1: "ecdsa-sha2-nistp256",
        SECP384R1: "ecdsa-sha2-nistp384",
    }
    _CURVE_NAMES = [curve.name for curve in _SCHEME_FOR_CURVE]

except ImportError:
    CRYPTO_IMPORT_ERROR = "'cryptography' required"

PYKCS11_IMPORT_ERROR = None
try:
    from PyKCS11 import PyKCS11

    # TODO: Don't hardcode schemes
    _MECHANISM_FOR_SCHEME = {
        "ecdsa-sha2-nistp256": PyKCS11.Mechanism(PyKCS11.CKM_ECDSA_SHA256),
        "ecdsa-sha2-nistp384": PyKCS11.Mechanism(PyKCS11.CKM_ECDSA_SHA384),
    }

except ImportError:
    PYKCS11_IMPORT_ERROR = "'PyKCS11' required"

ASN1_IMPORT_ERROR = None
try:
    from asn1crypto.keys import (  # pylint: disable=import-error
        ECDomainParameters,
        ECPoint,
    )
except ImportError:
    ASN1_IMPORT_ERROR = "'asn1crypto' required"
# pylint: enable=wrong-import-position


_PYKCS11LIB = None


def PYKCS11LIB():
    """Pseudo-singleton to load shared library using PYKCS11LIB envvar only once."""
    global _PYKCS11LIB  # pylint: disable=global-statement
    if _PYKCS11LIB is None:
        _PYKCS11LIB = PyKCS11.PyKCS11Lib()
        _PYKCS11LIB.load()

    return _PYKCS11LIB


class HSMSigner(Signer):
    """Hardware Security Module (HSM) Signer.

    Supports signing schemes "ecdsa-sha2-nistp256" and "ecdsa-sha2-nistp384".

    HSMSigner uses the first token it finds, if multiple tokens are available. They can
    be instantiated with Signer.from_priv_key_uri(). These private key URI schemes are
    supported:

    * "hsm:":
      Sign with key on PIV digital signature slot 9c.

    Usage::

        # sign with PIV slot 9c, verify with existing public key
        def pin_handler(secret: str) -> str:
            return getpass(f"Enter {secret}: ")

        signer = Signer.from_priv_key_uri("hsm:", public_key, pin_handler)
        sig = signer.sign(b"DATA")

        public_key.verify_signature(sig, b"DATA")

    Arguments:
        hsm_keyid: Key identifier on the token.
        public_key: The related public key instance.
        pin_handler: A function that returns the HSM user login pin, needed for
                signing. It receives the string argument "pin".

    Raises:
        UnsupportedLibraryError: ``PyKCS11`` and ``cryptography`` libraries not found.
        ValueError: ``public_key.scheme`` not supported.
    """

    # See Yubico docs for PKCS keyid to PIV slot mapping
    # https://developers.yubico.com/PIV/Introduction/Certificate_slots.html
    # https://developers.yubico.com/yubico-piv-tool/YKCS11/
    SCHEME_KEYID = 2
    SCHEME = "hsm"
    SECRETS_HANDLER_MSG = "pin"

    def __init__(
        self, hsm_keyid: int, public_key: Key, pin_handler: SecretsHandler
    ):
        if CRYPTO_IMPORT_ERROR:
            raise UnsupportedLibraryError(CRYPTO_IMPORT_ERROR)

        if PYKCS11_IMPORT_ERROR:
            raise UnsupportedLibraryError(PYKCS11_IMPORT_ERROR)

        if public_key.scheme not in _MECHANISM_FOR_SCHEME:
            raise ValueError(f"unsupported scheme {public_key.scheme}")

        self._mechanism = _MECHANISM_FOR_SCHEME[public_key.scheme]
        self.hsm_keyid = hsm_keyid
        self.public_key = public_key
        self.pin_handler = pin_handler

    @staticmethod
    @contextmanager
    def _default_session():
        """Context manager to handle default HSM session on reader slot 1."""
        lib = PYKCS11LIB()
        slots = lib.getSlotList(tokenPresent=True)
        if not slots:
            raise ValueError("could not find token")

        session = lib.openSession(slots[0])
        try:
            yield session

        finally:
            session.closeSession()

    @classmethod
    def _find_key(
        cls,
        session: "PyKCS11.Session",
        keyid: int,
        key_type: Optional[int] = None,
    ) -> int:
        """Find ecdsa key on HSM."""
        if key_type is None:
            key_type = PyKCS11.CKO_PUBLIC_KEY

        keys = session.findObjects(
            [
                (PyKCS11.CKA_CLASS, key_type),
                (PyKCS11.CKA_KEY_TYPE, PyKCS11.CKK_ECDSA),
                (PyKCS11.CKA_ID, (keyid,)),
            ]
        )
        if not keys:
            raise ValueError(f"could not find {KEY_TYPE_ECDSA} key for {keyid}")

        if len(keys) > 1:
            raise ValueError(
                f"found more than one {KEY_TYPE_ECDSA} key for {keyid}"
            )

        return keys[0]

    @classmethod
    def _find_key_values(
        cls, session: "PyKCS11.Session", keyid: int
    ) -> Tuple["ECDomainParameters", bytes]:
        """Find ecdsa public key values on HSM."""
        key = cls._find_key(session, keyid)
        params, point = session.getAttributeValue(
            key, [PyKCS11.CKA_EC_PARAMS, PyKCS11.CKA_EC_POINT]
        )
        return ECDomainParameters.load(bytes(params)), bytes(point)

    @classmethod
    def import_(cls, hsm_keyid: Optional[int] = None) -> Tuple[str, SSlibKey]:
        """Import public key and signer details from HSM.

        Returns a private key URI (for Signer.from_priv_key_uri()) and a public
        key. import_() should be called once and the returned URI and public
        key should be stored for later use.

        Arguments:
            hsm_keyid: Key identifier on the token. Default is 2 (meaning PIV key slot 9c).

        Raises:
            UnsupportedLibraryError: ``PyKCS11``, ``cryptography`` or ``asn1crypto``
                    libraries not found.
            ValueError: No compatible key for ``hsm_keyid`` found on HSM.
            PyKCS11.PyKCS11Error: Various HSM communication errors.

        """
        if CRYPTO_IMPORT_ERROR:
            raise UnsupportedLibraryError(CRYPTO_IMPORT_ERROR)

        if PYKCS11_IMPORT_ERROR:
            raise UnsupportedLibraryError(PYKCS11_IMPORT_ERROR)

        if ASN1_IMPORT_ERROR:
            raise UnsupportedLibraryError(ASN1_IMPORT_ERROR)

        if hsm_keyid is None:
            hsm_keyid = cls.SCHEME_KEYID

        with cls._default_session() as session:
            params, point = cls._find_key_values(session, hsm_keyid)

        if params.chosen.native not in _CURVE_NAMES:
            raise ValueError(
                f"found key on {params.chosen.native}, should be on one of {_CURVE_NAMES}"
            )

        # Create PEM from key
        curve = get_curve_for_oid(ObjectIdentifier(params.chosen.dotted))
        public_pem = (
            EllipticCurvePublicKey.from_encoded_point(
                curve(), ECPoint().load(point).native
            )
            .public_bytes(
                serialization.Encoding.PEM,
                serialization.PublicFormat.SubjectPublicKeyInfo,
            )
            .decode()
        )

        keyval = {"public": public_pem}
        scheme = _SCHEME_FOR_CURVE[curve]
        keyid = _get_keyid(KEY_TYPE_ECDSA, scheme, keyval)
        key = SSlibKey(keyid, KEY_TYPE_ECDSA, scheme, keyval)

        return "hsm:", key

    @classmethod
    def from_priv_key_uri(
        cls,
        priv_key_uri: str,
        public_key: Key,
        secrets_handler: Optional[SecretsHandler] = None,
    ) -> "HSMSigner":
        if not isinstance(public_key, SSlibKey):
            raise ValueError(f"expected SSlibKey for {priv_key_uri}")

        uri = parse.urlparse(priv_key_uri)

        if uri.scheme != cls.SCHEME:
            raise ValueError(f"HSMSigner does not support {priv_key_uri}")

        if secrets_handler is None:
            raise ValueError("HSMSigner requires a secrets handler")

        return cls(cls.SCHEME_KEYID, public_key, secrets_handler)

    def sign(self, payload: bytes) -> Signature:
        """Signs payload with Hardware Security Module (HSM).

        Arguments:
            payload: bytes to be signed.

        Raises:
            ValueError: No compatible key for ``hsm_keyid`` found on HSM.
            PyKCS11.PyKCS11Error: Various HSM communication errors.

        Returns:
            Signature.
        """
        pin = self.pin_handler(self.SECRETS_HANDLER_MSG)
        with self._default_session() as session:
            session.login(pin)
            key = self._find_key(
                session, self.hsm_keyid, PyKCS11.CKO_PRIVATE_KEY
            )
            signature = session.sign(key, payload, self._mechanism)
            session.logout()

        # The PKCS11 signature octets correspond to the concatenation of the ECDSA
        # values r and s, both represented as an octet string of equal length of at
        # most nLen with the most significant byte first (i.e. big endian)
        # https://docs.oasis-open.org/pkcs11/pkcs11-curr/v3.0/cs01/pkcs11-curr-v3.0-cs01.html#_Toc30061178
        r_s_len = int(len(signature) / 2)
        r = int.from_bytes(signature[:r_s_len], byteorder="big")
        s = int.from_bytes(signature[r_s_len:], byteorder="big")

        # Create an ASN.1 encoded Dss-Sig-Value to be used with pyca/cryptography
        dss_sig_value = binascii.hexlify(encode_dss_signature(r, s)).decode(
            "ascii"
        )

        return Signature(self.public_key.keyid, dss_sig_value)



"""Key interface and the default implementations"""
import logging
from abc import ABCMeta, abstractmethod
from typing import Any, Dict, Optional, Tuple, Type

import securesystemslib.keys as sslib_keys
from securesystemslib import exceptions
from securesystemslib.signer._signature import Signature

logger = logging.getLogger(__name__)

# NOTE Key dispatch table is defined here so it's usable by Key,
# but is populated in __init__.py (and can be appended by users).
KEY_FOR_TYPE_AND_SCHEME: Dict[Tuple[str, str], Type] = {}


class Key(metaclass=ABCMeta):
    """Abstract class representing the public portion of a key.

    *All parameters named below are not just constructor arguments but also
    instance attributes.*

    Args:
        keyid: Key identifier that is unique within the metadata it is used in.
            Keyid is not verified to be the hash of a specific representation
            of the key.
        keytype: Key type, e.g. "rsa", "ed25519" or "ecdsa-sha2-nistp256".
        scheme: Signature scheme. For example:
            "rsassa-pss-sha256", "ed25519", and "ecdsa-sha2-nistp256".
        keyval: Opaque key content
        unrecognized_fields: Dictionary of all attributes that are not managed
            by Securesystemslib

    Raises:
        TypeError: Invalid type for an argument.
    """

    def __init__(
        self,
        keyid: str,
        keytype: str,
        scheme: str,
        keyval: Dict[str, Any],
        unrecognized_fields: Optional[Dict[str, Any]] = None,
    ):
        if not all(
            isinstance(at, str) for at in [keyid, keytype, scheme]
        ) or not isinstance(keyval, dict):
            raise TypeError("Unexpected Key attributes types!")
        self.keyid = keyid
        self.keytype = keytype
        self.scheme = scheme
        self.keyval = keyval

        if unrecognized_fields is None:
            unrecognized_fields = {}

        self.unrecognized_fields = unrecognized_fields

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Key):
            return False

        return (
            self.keyid == other.keyid
            and self.keytype == other.keytype
            and self.scheme == other.scheme
            and self.keyval == other.keyval
            and self.unrecognized_fields == other.unrecognized_fields
        )

    @classmethod
    @abstractmethod
    def from_dict(cls, keyid: str, key_dict: Dict[str, Any]) -> "Key":
        """Creates ``Key`` object from a serialization dict

        Key implementations must override this factory constructor that is used
        as a deserialization helper.

        Users should call Key.from_dict(): it dispatches to the actual subclass
        implementation based on supported keys in KEY_FOR_TYPE_AND_SCHEME.

        Raises:
            KeyError, TypeError: Invalid arguments.
        """
        keytype = key_dict.get("keytype")
        scheme = key_dict.get("scheme")
        if (keytype, scheme) not in KEY_FOR_TYPE_AND_SCHEME:
            raise ValueError(f"Unsupported public key {keytype}/{scheme}")

        # NOTE: Explicitly not checking the keytype and scheme types to allow
        # intoto to use (None,None) to lookup GPGKey, see issue #450
        key_impl = KEY_FOR_TYPE_AND_SCHEME[(keytype, scheme)]  # type: ignore
        return key_impl.from_dict(keyid, key_dict)

    @abstractmethod
    def to_dict(self) -> Dict[str, Any]:
        """Returns a serialization dict.

        Key implementations must override this serialization helper.
        """
        raise NotImplementedError

    @abstractmethod
    def verify_signature(self, signature: Signature, data: bytes) -> None:
        """Raises if verification of signature over data fails.

        Args:
            signature: Signature object.
            data: Payload bytes.

        Raises:
            UnverifiedSignatureError: Failed to verify signature.
            VerificationError: Signature verification process error. If you
                are only interested in the verify result, just handle
                UnverifiedSignatureError: it contains VerificationError as well
        """
        raise NotImplementedError


class SSlibKey(Key):
    """Key implementation for RSA, Ed25519, ECDSA and Sphincs keys"""

    def to_securesystemslib_key(self) -> Dict[str, Any]:
        """Internal helper, returns a classic securesystemslib keydict"""
        return {
            "keyid": self.keyid,
            "keytype": self.keytype,
            "scheme": self.scheme,
            "keyval": self.keyval,
        }

    @classmethod
    def from_securesystemslib_key(cls, key_dict: Dict[str, Any]) -> "SSlibKey":
        """Constructor from classic securesystemslib keydict"""
        # ensure possible private keys are not included in keyval
        return SSlibKey(
            key_dict["keyid"],
            key_dict["keytype"],
            key_dict["scheme"],
            {"public": key_dict["keyval"]["public"]},
        )

    @classmethod
    def from_dict(cls, keyid: str, key_dict: Dict[str, Any]) -> "SSlibKey":
        keytype = key_dict.pop("keytype")
        scheme = key_dict.pop("scheme")
        keyval = key_dict.pop("keyval")

        if "public" not in keyval or not isinstance(keyval["public"], str):
            raise ValueError(f"public key string required for scheme {scheme}")

        # All fields left in the key_dict are unrecognized.
        return cls(keyid, keytype, scheme, keyval, key_dict)

    def to_dict(self) -> Dict[str, Any]:
        return {
            "keytype": self.keytype,
            "scheme": self.scheme,
            "keyval": self.keyval,
            **self.unrecognized_fields,
        }

    def verify_signature(self, signature: Signature, data: bytes) -> None:
        try:
            if not sslib_keys.verify_signature(
                self.to_securesystemslib_key(),
                signature.to_dict(),
                data,
            ):
                raise exceptions.UnverifiedSignatureError(
                    f"Failed to verify signature by {self.keyid}"
                )
        except (
            exceptions.CryptoError,
            exceptions.FormatError,
            exceptions.UnsupportedAlgorithmError,
        ) as e:
            logger.info("Key %s failed to verify sig: %s", self.keyid, str(e))
            raise exceptions.VerificationError(
                f"Unknown failure to verify signature by {self.keyid}"
            ) from e



#!/usr/bin/env python

"""
<Program Name>
  test_rsa_keys.py

<Author>
  Vladimir Diaz <vladimir.v.diaz@gmail.com>

<Started>
  June 3, 2015.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Test cases for 'rsa_keys.py'.
"""

import unittest

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import padding
from cryptography.hazmat.primitives.serialization import load_pem_private_key

import securesystemslib.exceptions
import securesystemslib.formats
import securesystemslib.hash
import securesystemslib.keys
import securesystemslib.rsa_keys

(
    public_rsa,
    private_rsa,
) = securesystemslib.rsa_keys.generate_rsa_public_and_private()
FORMAT_ERROR_MSG = (
    "securesystemslib.exceptions.FormatError raised.  Check object's format."
)


class TestRSA_keys(
    unittest.TestCase
):  # pylint: disable=missing-class-docstring,invalid-name
    def setUp(self):
        pass

    def test_generate_rsa_public_and_private(self):
        pub, priv = securesystemslib.rsa_keys.generate_rsa_public_and_private()

        # Check format of 'pub' and 'priv'.
        self.assertEqual(
            None,
            securesystemslib.formats.PEMRSA_SCHEMA.check_match(pub),
            FORMAT_ERROR_MSG,
        )
        self.assertEqual(
            None,
            securesystemslib.formats.PEMRSA_SCHEMA.check_match(priv),
            FORMAT_ERROR_MSG,
        )

        # Check for an invalid "bits" argument.  bits >= 2048.
        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.rsa_keys.generate_rsa_public_and_private,
            1024,
        )

        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.rsa_keys.generate_rsa_public_and_private,
            "2048",
        )

    def test_create_rsa_signature(self):
        global private_rsa  # pylint: disable=global-variable-not-assigned
        global public_rsa  # pylint: disable=global-variable-not-assigned
        data = "The quick brown fox jumps over the lazy dog".encode("utf-8")

        for rsa_scheme in securesystemslib.keys.RSA_SIGNATURE_SCHEMES:
            signature, scheme = securesystemslib.rsa_keys.create_rsa_signature(
                private_rsa, data, rsa_scheme
            )

            # Verify format of returned values.
            self.assertNotEqual(None, signature)
            self.assertEqual(
                None,
                securesystemslib.formats.RSA_SCHEME_SCHEMA.check_match(scheme),
                FORMAT_ERROR_MSG,
            )
            self.assertEqual(rsa_scheme, scheme)

            # Check for improperly formatted arguments.
            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.create_rsa_signature,
                123,
                data,
            )

            # Check for an unset private key.
            self.assertRaises(
                ValueError,
                securesystemslib.rsa_keys.create_rsa_signature,
                "",
                data,
            )

            # Check for an invalid PEM.
            self.assertRaises(
                securesystemslib.exceptions.CryptoError,
                securesystemslib.rsa_keys.create_rsa_signature,
                "123",
                data,
            )

            # Check for invalid 'data'.
            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.create_rsa_signature,
                private_rsa,
                "",
            )

            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.create_rsa_signature,
                private_rsa,
                123,
            )

            # Check for a missing private key.
            self.assertRaises(
                securesystemslib.exceptions.CryptoError,
                securesystemslib.rsa_keys.create_rsa_signature,
                public_rsa,
                data,
            )

            # Check for a TypeError by attempting to create a signature with an
            # encrypted key.
            encrypted_pem = securesystemslib.rsa_keys.create_rsa_encrypted_pem(
                private_rsa, "pw"
            )
            self.assertRaises(
                securesystemslib.exceptions.CryptoError,
                securesystemslib.rsa_keys.create_rsa_signature,
                encrypted_pem,
                data,
            )

    def test_verify_rsa_signature(self):
        global public_rsa  # pylint: disable=global-variable-not-assigned
        global private_rsa  # pylint: disable=global-variable-not-assigned
        data = "The quick brown fox jumps over the lazy dog".encode("utf-8")

        for rsa_scheme in securesystemslib.keys.RSA_SIGNATURE_SCHEMES:
            signature, scheme = securesystemslib.rsa_keys.create_rsa_signature(
                private_rsa, data, rsa_scheme
            )

            valid_signature = securesystemslib.rsa_keys.verify_rsa_signature(
                signature, scheme, public_rsa, data
            )
            self.assertEqual(True, valid_signature)

            # Check for an invalid public key.
            self.assertRaises(
                securesystemslib.exceptions.CryptoError,
                securesystemslib.rsa_keys.verify_rsa_signature,
                signature,
                scheme,
                private_rsa,
                data,
            )

            # Check for improperly formatted arguments.
            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.verify_rsa_signature,
                signature,
                123,
                public_rsa,
                data,
            )

            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.verify_rsa_signature,
                signature,
                scheme,
                123,
                data,
            )

            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.verify_rsa_signature,
                123,
                scheme,
                public_rsa,
                data,
            )

            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.verify_rsa_signature,
                signature,
                "invalid_scheme",
                public_rsa,
                data,
            )

            # Check for invalid 'signature' and 'data' arguments.
            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.rsa_keys.verify_rsa_signature,
                signature,
                scheme,
                public_rsa,
                123,
            )

            self.assertEqual(
                False,
                securesystemslib.rsa_keys.verify_rsa_signature(
                    signature, scheme, public_rsa, b"mismatched data"
                ),
            )

            (
                mismatched_signature,
                scheme,
            ) = securesystemslib.rsa_keys.create_rsa_signature(
                private_rsa, b"mismatched data"
            )

            self.assertEqual(
                False,
                securesystemslib.rsa_keys.verify_rsa_signature(
                    mismatched_signature, scheme, public_rsa, data
                ),
            )

    def test_verify_rsa_pss_different_salt_lengths(self):
        rsa_scheme = "rsassa-pss-sha256"
        data = "The ancients say, salt length does not matter that much".encode(
            "utf-8"
        )

        private_key = load_pem_private_key(
            private_rsa.encode("utf-8"),
            password=None,
            backend=default_backend(),
        )
        digest = securesystemslib.hash.digest_from_rsa_scheme(
            rsa_scheme, "pyca_crypto"
        )

        # Make sure digest size and max salt length are not accidentally the same
        self.assertNotEqual(
            digest.algorithm.digest_size,
            padding.calculate_max_pss_salt_length(
                private_key, digest.algorithm
            ),
        )

        # Sign with max salt length (briefly available in sslib v0.24.0):
        max_salt_sig = private_key.sign(
            data,
            padding.PSS(
                mgf=padding.MGF1(digest.algorithm),
                salt_length=padding.PSS.MAX_LENGTH,
            ),
            digest.algorithm,
        )

        # Sign with salt length == digest length
        fix_salt_sig, _ = securesystemslib.rsa_keys.create_rsa_signature(
            private_rsa, data
        )

        # Verification infers salt length automatically and so works for both
        for signature in (max_salt_sig, fix_salt_sig):
            verified = securesystemslib.rsa_keys.verify_rsa_signature(
                signature, rsa_scheme, public_rsa, data
            )
            self.assertTrue(verified)

    def test_create_rsa_encrypted_pem(self):
        global public_rsa  # pylint: disable=global-variable-not-assigned
        global private_rsa  # pylint: disable=global-variable-not-assigned

        encrypted_pem = securesystemslib.rsa_keys.create_rsa_encrypted_pem(
            private_rsa, "password"
        )
        self.assertTrue(
            securesystemslib.formats.PEMRSA_SCHEMA.matches(encrypted_pem)
        )

        # Test for invalid private key (via PEM).
        self.assertRaises(
            securesystemslib.exceptions.CryptoError,
            securesystemslib.rsa_keys.create_rsa_encrypted_pem,
            public_rsa,
            "password",
        )

        # Test for invalid arguments.
        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.rsa_keys.create_rsa_encrypted_pem,
            public_rsa,
            123,
        )

        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.rsa_keys.create_rsa_encrypted_pem,
            123,
            "password",
        )

        self.assertRaises(
            ValueError,
            securesystemslib.rsa_keys.create_rsa_encrypted_pem,
            "",
            "password",
        )

    def test_create_rsa_public_and_private_from_pem(self):
        global public_rsa  # pylint: disable=global-variable-not-assigned
        global private_rsa  # pylint: disable=global-variable-not-assigned

        (
            public,
            private,
        ) = securesystemslib.rsa_keys.create_rsa_public_and_private_from_pem(
            private_rsa
        )

        self.assertTrue(securesystemslib.formats.PEMRSA_SCHEMA.matches(public))
        self.assertTrue(securesystemslib.formats.PEMRSA_SCHEMA.matches(private))

        self.assertRaises(
            securesystemslib.exceptions.CryptoError,
            securesystemslib.rsa_keys.create_rsa_public_and_private_from_pem,
            public_rsa,
        )

    def test_encrypt_key(self):
        global public_rsa  # pylint: disable=global-variable-not-assigned
        global private_rsa  # pylint: disable=global-variable-not-assigned

        key_object = {
            "keytype": "rsa",
            "scheme": "rsassa-pss-sha256",
            "keyid": "1223",
            "keyval": {"public": public_rsa, "private": private_rsa},
        }

        encrypted_key = securesystemslib.rsa_keys.encrypt_key(
            key_object, "password"
        )
        self.assertTrue(
            securesystemslib.formats.ENCRYPTEDKEY_SCHEMA.matches(encrypted_key)
        )

        key_object["keyval"]["private"] = ""
        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.rsa_keys.encrypt_key,
            key_object,
            "password",
        )

    def test_decrypt_key(self):

        # Test for valid arguments.
        global public_rsa  # pylint: disable=global-variable-not-assigned
        global private_rsa  # pylint: disable=global-variable-not-assigned
        passphrase = "pw"

        rsa_key = {
            "keytype": "rsa",
            "scheme": "rsassa-pss-sha256",
            "keyid": "d62247f817883f593cf6c66a5a55292488d457bcf638ae03207dbbba9dbe457d",
            "keyval": {"public": public_rsa, "private": private_rsa},
        }

        encrypted_rsa_key = securesystemslib.rsa_keys.encrypt_key(
            rsa_key, passphrase
        )

        _ = securesystemslib.rsa_keys.decrypt_key(encrypted_rsa_key, passphrase)

        # Test for invalid arguments.
        self.assertRaises(
            securesystemslib.exceptions.CryptoError,
            securesystemslib.rsa_keys.decrypt_key,
            "bad",
            passphrase,
        )

        # Test for invalid encrypted content (i.e., invalid hmac and ciphertext.)
        encryption_delimiter = (
            securesystemslib.rsa_keys._ENCRYPTION_DELIMITER  # pylint: disable=protected-access
        )
        salt, iterations, hmac, iv, ciphertext = encrypted_rsa_key.split(
            encryption_delimiter
        )

        # Set an invalid hmac.  The decryption routine sould raise a
        # securesystemslib.exceptions.CryptoError exception because 'hmac' does not
        # match the hmac calculated by the decryption routine.
        bad_hmac = "12345abcd"
        invalid_encrypted_rsa_key = (
            salt
            + encryption_delimiter
            + iterations
            + encryption_delimiter
            + bad_hmac
            + encryption_delimiter
            + iv
            + encryption_delimiter
            + ciphertext
        )

        self.assertRaises(
            securesystemslib.exceptions.CryptoError,
            securesystemslib.rsa_keys.decrypt_key,
            invalid_encrypted_rsa_key,
            passphrase,
        )

        # Test for invalid 'ciphertext'
        bad_ciphertext = "12345abcde"
        invalid_encrypted_rsa_key = (
            salt
            + encryption_delimiter
            + iterations
            + encryption_delimiter
            + hmac
            + encryption_delimiter
            + iv
            + encryption_delimiter
            + bad_ciphertext
        )

        self.assertRaises(
            securesystemslib.exceptions.CryptoError,
            securesystemslib.rsa_keys.decrypt_key,
            invalid_encrypted_rsa_key,
            passphrase,
        )


# Run the unit tests.
if __name__ == "__main__":
    unittest.main()



"""
<Program Name>
  util.py

<Author>
  Konstantin Andrianov

<Started>
  March 24, 2012.  Derived from original util.py written by Geremy Condra.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Provides utility services.  This module supplies utility functions such as:
  get_file_details() that computes the length and hash of a file, import_json
  that tries to import a working json module, load_json_* functions, etc.
"""

import json
import logging
import os
from typing import IO, Any, Dict, List, Optional, Sequence, Tuple, Union

from securesystemslib import exceptions, formats
from securesystemslib.hash import digest_fileobject
from securesystemslib.storage import FilesystemBackend, StorageBackendInterface

logger = logging.getLogger(__name__)


def get_file_details(  # pylint: disable=dangerous-default-value
    filepath: str,
    hash_algorithms: List[str] = ["sha256"],
    storage_backend: Optional[StorageBackendInterface] = None,
) -> Tuple[int, Dict[str, str]]:
    """
    <Purpose>
      To get file's length and hash information.  The hash is computed using the
      sha256 algorithm.  This function is used in the signerlib.py and updater.py
      modules.

    <Arguments>
      filepath:
        Absolute file path of a file.

      hash_algorithms:
        A list of hash algorithms with which the file's hash should be computed.
        Defaults to ['sha256']

      storage_backend:
        An object which implements
        securesystemslib.storage.StorageBackendInterface. When no object is
        passed a FilesystemBackend will be instantiated and used.

    <Exceptions>
      securesystemslib.exceptions.FormatError: If hash of the file does not match
      HASHDICT_SCHEMA.

      securesystemslib.exceptions.StorageError: The file at "filepath" cannot be
      opened or found.

    <Returns>
      A tuple (length, hashes) describing 'filepath'.
    """

    # Making sure that the format of 'filepath' is a path string.
    # 'securesystemslib.exceptions.FormatError' is raised on incorrect format.
    formats.PATH_SCHEMA.check_match(filepath)
    formats.HASHALGORITHMS_SCHEMA.check_match(hash_algorithms)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    file_length = get_file_length(filepath, storage_backend)
    file_hashes = get_file_hashes(filepath, hash_algorithms, storage_backend)

    return file_length, file_hashes


def get_file_hashes(  # pylint: disable=dangerous-default-value
    filepath: str,
    hash_algorithms: List[str] = ["sha256"],
    storage_backend: Optional[StorageBackendInterface] = None,
) -> Dict[str, str]:
    """
    <Purpose>
      Compute hash(es) of the file at filepath using each of the specified
      hash algorithms. If no algorithms are specified, then the hash is
      computed using the SHA-256 algorithm.

    <Arguments>
      filepath:
        Absolute file path of a file.

      hash_algorithms:
        A list of hash algorithms with which the file's hash should be computed.
        Defaults to ['sha256']

      storage_backend:
        An object which implements
        securesystemslib.storage.StorageBackendInterface. When no object is
        passed a FilesystemBackend will be instantiated and used.

    <Exceptions>
      securesystemslib.exceptions.FormatError: If hash of the file does not match
      HASHDICT_SCHEMA.

      securesystemslib.exceptions.StorageError: The file at "filepath" cannot be
      opened or found.

    <Returns>
      A dictionary conforming to securesystemslib.formats.HASHDICT_SCHEMA
      containing information about the hashes of the file at "filepath".
    """

    # Making sure that the format of 'filepath' is a path string.
    # 'securesystemslib.exceptions.FormatError' is raised on incorrect format.
    formats.PATH_SCHEMA.check_match(filepath)
    formats.HASHALGORITHMS_SCHEMA.check_match(hash_algorithms)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    file_hashes = {}

    with storage_backend.get(filepath) as fileobj:
        # Obtaining hash of the file.
        for algorithm in hash_algorithms:
            digest_object = digest_fileobject(fileobj, algorithm)
            file_hashes.update({algorithm: digest_object.hexdigest()})

    # Performing a format check to ensure 'file_hash' corresponds HASHDICT_SCHEMA.
    # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
    formats.HASHDICT_SCHEMA.check_match(file_hashes)

    return file_hashes


def get_file_length(
    filepath: str, storage_backend: Optional[StorageBackendInterface] = None
) -> int:
    """
    <Purpose>
      To get file's length information.

    <Arguments>
      filepath:
        Absolute file path of a file.

      storage_backend:
        An object which implements
        securesystemslib.storage.StorageBackendInterface. When no object is
        passed a FilesystemBackend will be instantiated and used.

    <Exceptions>
      securesystemslib.exceptions.StorageError: The file at "filepath" cannot be
      opened or found.

    <Returns>
      The length, in bytes, of the file at 'filepath'.
    """

    # Making sure that the format of 'filepath' is a path string.
    # 'securesystemslib.exceptions.FormatError' is raised on incorrect format.
    formats.PATH_SCHEMA.check_match(filepath)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    return storage_backend.getsize(filepath)


def persist_temp_file(
    temp_file: IO,
    persist_path: str,
    storage_backend: Optional[StorageBackendInterface] = None,
    should_close: bool = True,
    restrict: bool = False,
) -> None:
    """
    <Purpose>
      Copies 'temp_file' (a file like object) to a newly created non-temp file at
      'persist_path'.

    <Arguments>
      temp_file:
        File object to persist, typically a file object returned by one of the
        interfaces in the tempfile module of the standard library.

      persist_path:
        File path to create the persistent file in.

      storage_backend:
        An object which implements
        securesystemslib.storage.StorageBackendInterface. When no object is
        passed a FilesystemBackend will be instantiated and used.

      should_close:
        A boolean indicating whether the file should be closed after it has been
        persisted. Default is True, the file is closed.

      restrict:
        A boolean indicating whether the file should have restricted privileges.
        What evactly counts as restricted privileges is an implementation detail
        of the backing StorageBackendInterface implementation.

    <Exceptions>
      securesystemslib.exceptions.StorageError: If file cannot be written.

    <Return>
      None.
    """

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    storage_backend.put(temp_file, persist_path, restrict=restrict)

    if should_close:
        temp_file.close()


def ensure_parent_dir(
    filename: str, storage_backend: Optional[StorageBackendInterface] = None
) -> None:
    """
    <Purpose>
      To ensure existence of the parent directory of 'filename'.  If the parent
      directory of 'name' does not exist, create it.

      Example: If 'filename' is '/a/b/c/d.txt', and only the directory '/a/b/'
      exists, then directory '/a/b/c/d/' will be created.

    <Arguments>
      filename:
        A path string.

      storage_backend:
        An object which implements
        securesystemslib.storage.StorageBackendInterface. When no object is
        passed a FilesystemBackend will be instantiated and used.

    <Exceptions>
      securesystemslib.exceptions.FormatError: If 'filename' is improperly
      formatted.
      securesystemslib.exceptions.StorageError: When folder cannot be created.

    <Side Effects>
      A directory is created whenever the parent directory of 'filename' does not
      exist.

    <Return>
      None.
    """

    # Ensure 'filename' corresponds to 'PATH_SCHEMA'.
    # Raise 'securesystemslib.exceptions.FormatError' on a mismatch.
    formats.PATH_SCHEMA.check_match(filename)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    # Split 'filename' into head and tail, check if head exists.
    directory = os.path.split(filename)[0]

    # Check for cases where filename is without directory like 'file.txt'
    # and as a result directory is an empty string
    if directory:
        storage_backend.create_folder(directory)


def file_in_confined_directories(
    filepath: str, confined_directories: Sequence[str]
) -> bool:
    """
    <Purpose>
      Check if the directory containing 'filepath' is in the list/tuple of
      'confined_directories'.

    <Arguments>
      filepath:
        A string representing the path of a file.  The following example path
        strings are viewed as files and not directories: 'a/b/c', 'a/b/c.txt'.

      confined_directories:
        A sequence (such as list, or tuple) of directory strings.

    <Exceptions>
     securesystemslib.exceptions.FormatError: On incorrect format of the input.

    <Return>
      Boolean.  True, if path is either the empty string
      or in 'confined_paths'; False, otherwise.
    """

    # Do the arguments have the correct format?
    # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
    formats.PATH_SCHEMA.check_match(filepath)
    formats.NAMES_SCHEMA.check_match(confined_directories)

    for confined_directory in confined_directories:
        # The empty string (arbitrarily chosen) signifies the client is confined
        # to all directories and subdirectories.  No need to check 'filepath'.
        if confined_directory == "":
            return True

        # Normalized paths needed, to account for up-level references, etc.
        # callers have the option of setting the list of directories in
        # 'confined_directories'.
        filepath = os.path.normpath(filepath)
        confined_directory = os.path.normpath(confined_directory)

        # A caller may restrict himself to specific directories on the
        # remote repository.  The list of paths in 'confined_path', not including
        # each path's subdirectories, are the only directories the client will
        # download targets from.
        if os.path.dirname(filepath) == confined_directory:
            return True

    return False


def load_json_string(data: Union[str, bytes]) -> Any:
    """
    <Purpose>
      Deserialize 'data' (JSON string) to a Python object.

    <Arguments>
      data:
        A JSON string.

    <Exceptions>
      securesystemslib.exceptions.Error, if 'data' cannot be deserialized to a
      Python object.

    <Side Effects>
      None.

    <Returns>
      Deserialized object.  For example, a dictionary.
    """

    deserialized_object = None

    try:
        deserialized_object = json.loads(data)

    except TypeError:
        message = "Invalid JSON string: " + repr(data)
        raise exceptions.Error(message)  # pylint: disable=raise-missing-from

    except ValueError:
        message = "Cannot deserialize to a Python object: " + repr(data)
        raise exceptions.Error(message)  # pylint: disable=raise-missing-from

    else:
        return deserialized_object


def load_json_file(
    filepath: str, storage_backend: Optional[StorageBackendInterface] = None
) -> Any:
    """
    <Purpose>
      Deserialize a JSON object from a file containing the object.

    <Arguments>
      filepath:
        Absolute path of JSON file.

      storage_backend:
        An object which implements
        securesystemslib.storage.StorageBackendInterface. When no object is
        passed a FilesystemBackend will be instantiated and used.

    <Exceptions>
      securesystemslib.exceptions.FormatError: If 'filepath' is improperly
      formatted.

      securesystemslib.exceptions.Error: If 'filepath' cannot be deserialized to
      a Python object.

      securesystemslib.exceptions.StorageError: If file cannot be loaded.

      IOError in case of runtime IO exceptions.

    <Side Effects>
      None.

    <Return>
      Deserialized object.  For example, a dictionary.
    """

    # Making sure that the format of 'filepath' is a path string.
    # securesystemslib.exceptions.FormatError is raised on incorrect format.
    formats.PATH_SCHEMA.check_match(filepath)

    if storage_backend is None:
        storage_backend = FilesystemBackend()

    deserialized_object = None
    with storage_backend.get(filepath) as file_obj:
        raw_data = file_obj.read().decode("utf-8")

        try:
            deserialized_object = json.loads(raw_data)

        except (ValueError, TypeError):
            raise exceptions.Error(  # pylint: disable=raise-missing-from
                "Cannot deserialize to a" " Python object: " + filepath
            )

        else:
            return deserialized_object


def digests_are_equal(digest1: str, digest2: str) -> bool:
    """
    <Purpose>
      While protecting against timing attacks, compare the hexadecimal arguments
      and determine if they are equal.

    <Arguments>
      digest1:
        The first hexadecimal string value to compare.

      digest2:
        The second hexadecimal string value to compare.

    <Exceptions>
      securesystemslib.exceptions.FormatError: If the arguments are improperly
      formatted.

    <Side Effects>
      None.

    <Return>
      Return True if 'digest1' is equal to 'digest2', False otherwise.
    """

    # Ensure the arguments have the appropriate number of objects and object
    # types, and that all dict keys are properly named.
    # Raise 'securesystemslib.exceptions.FormatError' if there is a mismatch.
    formats.HEX_SCHEMA.check_match(digest1)
    formats.HEX_SCHEMA.check_match(digest2)

    if len(digest1) != len(digest2):
        return False

    are_equal = True

    for val1, val2 in zip(digest1, digest2):
        if val1 != val2:
            are_equal = False

    return are_equal



#!/usr/bin/env python

"""
<Program Name>
  test_gpg.py

<Author>
  Santiago Torres-Arias <santiago@nyu.edu>
  Lukas Puehringer <lukas.puehringer@nyu.edu>

<Started>
  Nov 15, 2017

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Test gpg/pgp-related functions.

"""

import os
import shutil
import tempfile
import unittest

# pylint: disable=wrong-import-position
from collections import OrderedDict
from copy import deepcopy
from unittest.mock import patch

import cryptography.hazmat.primitives.hashes as hashing
from cryptography.hazmat import backends
from cryptography.hazmat.primitives import serialization

from securesystemslib import exceptions, process
from securesystemslib.formats import ANY_PUBKEY_DICT_SCHEMA, GPG_PUBKEY_SCHEMA
from securesystemslib.gpg.common import (
    _assign_certified_key_info,
    _get_verified_subkeys,
    get_pubkey_bundle,
    parse_pubkey_bundle,
    parse_pubkey_payload,
    parse_signature_packet,
)
from securesystemslib.gpg.constants import (
    PACKET_TYPE_PRIMARY_KEY,
    PACKET_TYPE_SUB_KEY,
    PACKET_TYPE_USER_ATTR,
    PACKET_TYPE_USER_ID,
    SHA1,
    SHA256,
    SHA512,
    gpg_export_pubkey_command,
    have_gpg,
)
from securesystemslib.gpg.dsa import create_pubkey as dsa_create_pubkey
from securesystemslib.gpg.eddsa import ED25519_SIG_LENGTH
from securesystemslib.gpg.exceptions import (
    CommandError,
    KeyExpirationError,
    KeyNotFoundError,
    PacketParsingError,
    PacketVersionNotSupportedError,
    SignatureAlgorithmNotSupportedError,
)
from securesystemslib.gpg.functions import (
    create_signature,
    export_pubkey,
    export_pubkeys,
    verify_signature,
)
from securesystemslib.gpg.rsa import create_pubkey as rsa_create_pubkey
from securesystemslib.gpg.util import (
    Version,
    get_hashing_class,
    get_version,
    is_version_fully_supported,
    parse_packet_header,
    parse_subpacket_header,
)

# pylint: enable=wrong-import-position


class GPGTestUtils:
    """GPG Test utility class"""

    @staticmethod
    def ignore_not_found_error(
        function, path, exc_info
    ):  # pylint: disable=unused-argument,unused-argument
        """Callback that ignores FileNotFoundError"""
        _, error, _ = exc_info
        if not isinstance(error, FileNotFoundError):
            raise error


@unittest.skipIf(not have_gpg(), "gpg not found")
class TestUtil(unittest.TestCase):
    """Test util functions."""

    def test_version_utils_return_types(self):
        """Run dummy tests for coverage."""
        self.assertTrue(isinstance(get_version(), Version))
        self.assertTrue(isinstance(is_version_fully_supported(), bool))

    def test_version_utils_error(self):
        """Run dummy tests for coverage."""
        with patch(
            "securesystemslib.gpg.constants.have_gpg", return_value=False
        ):
            with self.assertRaises(exceptions.UnsupportedLibraryError):
                get_version()

    def test_get_hashing_class(self):
        # Assert return expected hashing class
        expected_hashing_class = [hashing.SHA1, hashing.SHA256, hashing.SHA512]
        for idx, hashing_id in enumerate([SHA1, SHA256, SHA512]):
            result = get_hashing_class(hashing_id)
            self.assertEqual(result, expected_hashing_class[idx])

        # Assert raises ValueError with non-supported hashing id
        with self.assertRaises(ValueError):
            get_hashing_class("bogus_hashing_id")

    def test_parse_packet_header(self):
        """Test parse_packet_header with manually crafted data."""
        data_list = [
            ## New format packet length with mock packet type 100001
            # one-octet length, header len: 2, body len: 0 to 191
            [0b01100001, 0],
            [0b01100001, 191],
            # two-octet length, header len: 3, body len: 192 to 8383
            [0b01100001, 192, 0],
            [0b01100001, 223, 255],
            # five-octet length, header len: 6, body len: 0 to 4,294,967,295
            [0b01100001, 255, 0, 0, 0, 0],
            [0b01100001, 255, 255, 255, 255, 255],
            ## Old format packet lengths with mock packet type 1001
            # one-octet length, header len: 2, body len: 0 to 255
            [0b00100100, 0],
            [0b00100100, 255],
            # two-octet length, header len: 3, body len: 0 to 65,535
            [0b00100101, 0, 0],
            [0b00100101, 255, 255],
            # four-octet length, header len: 5, body len: 0 to 4,294,967,295
            [0b00100110, 0, 0, 0, 0, 0],
            [0b00100110, 255, 255, 255, 255, 255],
        ]

        # packet_type | header_len | body_len | packet_len
        expected = [
            (33, 2, 0, 2),
            (33, 2, 191, 193),
            (33, 3, 192, 195),
            (33, 3, 8383, 8386),
            (33, 6, 0, 6),
            (33, 6, 4294967295, 4294967301),
            (9, 2, 0, 2),
            (9, 2, 255, 257),
            (9, 3, 0, 3),
            (9, 3, 65535, 65538),
            (9, 5, 0, 5),
            (9, 5, 4294967295, 4294967300),
        ]

        for idx, data in enumerate(data_list):
            result = parse_packet_header(bytearray(data))
            self.assertEqual(result, expected[idx])

        # New Format Packet Lengths with Partial Body Lengths range
        for second_octet in [224, 254]:
            with self.assertRaises(PacketParsingError):
                parse_packet_header(bytearray([0b01100001, second_octet]))

        # Old Format Packet Lengths with indeterminate length (length type 3)
        with self.assertRaises(PacketParsingError):
            parse_packet_header(bytearray([0b00100111]))

        # Get expected type
        parse_packet_header(bytearray([0b01100001, 0]), expected_type=33)

        # Raise with unexpected type
        with self.assertRaises(PacketParsingError):
            parse_packet_header(bytearray([0b01100001, 0]), expected_type=34)

    def test_parse_subpacket_header(self):
        """Test parse_subpacket_header with manually crafted data."""
        # All items until last item encode the length of the subpacket,
        # the last item encodes the mock subpacket type.
        data_list = [
            # length of length 1, subpacket length 0 to 191
            [0, 33],  # NOTE: Nonsense 0-length
            [191, 33],
            # # length of length 2, subpacket length 192 to 16,319
            [192, 0, 33],
            [254, 255, 33],
            # # length of length 5, subpacket length 0 to 4,294,967,295
            [255, 0, 0, 0, 0, 33],  # NOTE: Nonsense 0-length
            [255, 255, 255, 255, 255, 33],
        ]
        # packet_type | header_len | body_len | packet_len
        expected = [
            (33, 2, -1, 1),  # NOTE: Nonsense negative payload
            (33, 2, 190, 192),
            (33, 3, 191, 194),
            (33, 3, 16318, 16321),
            (33, 6, -1, 5),  # NOTE: Nonsense negative payload
            (33, 6, 4294967294, 4294967300),
        ]

        for idx, data in enumerate(data_list):
            result = parse_subpacket_header(bytearray(data))
            self.assertEqual(result, expected[idx])


@unittest.skipIf(not have_gpg(), "gpg not found")
class TestCommon(unittest.TestCase):
    """Test common functions of the securesystemslib.gpg module."""

    @classmethod
    def setUpClass(self):  # pylint: disable=bad-classmethod-argument
        gpg_keyring_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "gpg_keyrings", "rsa"
        )
        homearg = (
            "--homedir {}".format(  # pylint: disable=consider-using-f-string
                gpg_keyring_path
            ).replace("\\", "/")
        )

        # Load test raw public key bundle from rsa keyring, used to construct
        # erroneous gpg data in tests below.
        keyid = "F557D0FF451DEF45372591429EA70BD13D883381"
        cmd = gpg_export_pubkey_command(keyid=keyid, homearg=homearg)
        proc = process.run(cmd, stdout=process.PIPE, stderr=process.PIPE)
        self.raw_key_data = proc.stdout
        self.raw_key_bundle = parse_pubkey_bundle(self.raw_key_data)

        # Export pubkey bundle with expired key for key expiration tests
        keyid = "E8AC80C924116DABB51D4B987CB07D6D2C199C7C"
        cmd = gpg_export_pubkey_command(keyid=keyid, homearg=homearg)
        proc = process.run(cmd, stdout=process.PIPE, stderr=process.PIPE)
        self.raw_expired_key_bundle = parse_pubkey_bundle(proc.stdout)

    def test_parse_pubkey_payload_errors(self):
        """Test parse_pubkey_payload errors with manually crafted data."""
        # passed data | expected error | expected error message
        test_data = [
            (None, ValueError, "empty pubkey"),
            (
                bytearray([0x03]),
                PacketVersionNotSupportedError,
                "packet version '3' not supported",
            ),
            (
                bytearray([0x04, 0, 0, 0, 0, 255]),
                SignatureAlgorithmNotSupportedError,
                "Signature algorithm '255' not supported",
            ),
        ]

        for data, error, error_str in test_data:
            with self.assertRaises(error) as ctx:
                parse_pubkey_payload(data)
            self.assertTrue(error_str in str(ctx.exception))

    def test_parse_pubkey_bundle_errors(self):
        """Test parse_pubkey_bundle errors with manually crafted data partially
        based on real gpg key data (see self.raw_key_bundle)."""
        # Extract sample (legitimate) user ID packet and pass as first packet to
        # raise first packet must be primary key error
        user_id_packet = list(self.raw_key_bundle[PACKET_TYPE_USER_ID].keys())[
            0
        ]
        # Extract sample (legitimate) primary key packet and pass as first two
        # packets to raise unexpected second primary key error
        primary_key_packet = self.raw_key_bundle[PACKET_TYPE_PRIMARY_KEY][
            "packet"
        ]
        # Create incomplete packet to re-raise header parsing IndexError as
        # PacketParsingError
        incomplete_packet = bytearray([0b01111111])

        # passed data | expected error message
        test_data = [
            (None, "empty gpg data"),
            (user_id_packet, "must be a primary key"),
            (primary_key_packet + primary_key_packet, "Unexpected primary key"),
            (incomplete_packet, "index out of range"),
        ]
        for data, error_str in test_data:
            with self.assertRaises(PacketParsingError) as ctx:
                parse_pubkey_bundle(data)
            self.assertTrue(error_str in str(ctx.exception))

        # Create empty packet of unsupported type 66 (bit 0-5) and length 0 and
        # pass as second packet to provoke skipping of unsupported packet
        unsupported_packet = bytearray([0b01111111, 0])
        with patch("securesystemslib.gpg.common.log") as mock_log:
            parse_pubkey_bundle(primary_key_packet + unsupported_packet)
            self.assertTrue(
                "Ignoring gpg key packet '63'" in mock_log.info.call_args[0][0]
            )

    def test_parse_pubkey_bundle(self):
        """Assert presence of packets expected returned from `parse_pubkey_bundle`
    for specific test key). See
    ```
    gpg --homedir tests/gpg_keyrings/rsa/ --export 9EA70BD13D883381 | \
        gpg --list-packets
    ```
    """
        # Expect parsed primary key matching GPG_PUBKEY_SCHEMA
        self.assertTrue(
            GPG_PUBKEY_SCHEMA.matches(
                self.raw_key_bundle[PACKET_TYPE_PRIMARY_KEY]["key"]
            )
        )

        # Parse corresponding raw packet for comparison
        _, header_len, _, _ = parse_packet_header(
            self.raw_key_bundle[PACKET_TYPE_PRIMARY_KEY]["packet"]
        )

        # pylint: disable=unsubscriptable-object
        parsed_raw_packet = parse_pubkey_payload(
            bytearray(
                self.raw_key_bundle[PACKET_TYPE_PRIMARY_KEY]["packet"][
                    header_len:
                ]
            )
        )

        # And compare
        self.assertDictEqual(
            self.raw_key_bundle[PACKET_TYPE_PRIMARY_KEY]["key"],
            parsed_raw_packet,
        )

        # Expect one primary key signature (revocation signature)
        self.assertEqual(
            len(self.raw_key_bundle[PACKET_TYPE_PRIMARY_KEY]["signatures"]), 1
        )

        # Expect one User ID packet, one User Attribute packet and one Subkey,
        # each with correct data
        for _type in [
            PACKET_TYPE_USER_ID,
            PACKET_TYPE_USER_ATTR,
            PACKET_TYPE_SUB_KEY,
        ]:
            # Of each type there is only one packet
            self.assertTrue(len(self.raw_key_bundle[_type]) == 1)
            # The raw packet is stored as key in the per-packet type collection
            raw_packet = next(iter(self.raw_key_bundle[_type]))
            # Its values are the raw packets header and body length
            self.assertEqual(
                len(raw_packet),
                self.raw_key_bundle[_type][raw_packet]["header_len"]
                + self.raw_key_bundle[_type][raw_packet]["body_len"],
            )
            # and one self-signature
            self.assertEqual(
                len(self.raw_key_bundle[_type][raw_packet]["signatures"]), 1
            )

    def test_assign_certified_key_info_errors(self):
        """Test _assign_certified_key_info errors with manually crafted data
        based on real gpg key data (see self.raw_key_bundle)."""

        # Replace legitimate user certifacte with a bogus packet
        wrong_cert_bundle = deepcopy(self.raw_key_bundle)
        packet, packet_data = wrong_cert_bundle[PACKET_TYPE_USER_ID].popitem()
        packet_data["signatures"] = [bytearray([0b01111111, 0])]
        wrong_cert_bundle[PACKET_TYPE_USER_ID][packet] = packet_data

        # Replace primary key id with a non-associated keyid
        wrong_keyid_bundle = deepcopy(self.raw_key_bundle)
        wrong_keyid_bundle[PACKET_TYPE_PRIMARY_KEY]["key"][
            "keyid"
        ] = "8465A1E2E0FB2B40ADB2478E18FB3F537E0C8A17"

        # Remove a byte in user id packet to make signature verification fail
        invalid_cert_bundle = deepcopy(self.raw_key_bundle)
        packet, packet_data = invalid_cert_bundle[PACKET_TYPE_USER_ID].popitem()
        packet = packet[:-1]
        invalid_cert_bundle[PACKET_TYPE_USER_ID][packet] = packet_data

        test_data = [
            # Skip and log parse_signature_packet error
            (wrong_cert_bundle, "Expected packet 2, but got 63 instead"),
            # Skip and log signature packet that doesn't match primary key id
            (wrong_keyid_bundle, "Ignoring User ID certificate issued by"),
            # Skip and log invalid signature
            (invalid_cert_bundle, "Ignoring invalid User ID self-certificate"),
        ]

        for bundle, expected_msg in test_data:
            with patch("securesystemslib.gpg.common.log") as mock_log:
                _assign_certified_key_info(bundle)
                msg = str(mock_log.info.call_args[0][0])
                self.assertTrue(
                    expected_msg in msg,
                    "'{}' not in '{}'".format(  # pylint: disable=consider-using-f-string
                        expected_msg, msg
                    ),
                )

    def test_assign_certified_key_info_expiration(self):
        """Test assignment of key expiration date in
        gpg.common._assign_certified_key_info using real gpg data (with ambiguity
        resolution / prioritization).

        # FIXME: Below tests are missing proper assertions for which User ID
        self-certificate is considered for the expiration date. Reasons are:
        - gpg does not let you (easily) modify individual expiration dates of User
          IDs (changing one changes all), hence we cannot assert the chosen packet
          by the particular date
        -  _assign_certified_key_info first verifies all self-certificates and then
           only considers successfully verified ones, hence we cannot modify the
           certificate data, before passing it to _assign_certified_key_info

        IMO the best solution is a better separation of concerns, e.g. separate
        self-certificate verification and packet prioritization.

        """
        # Test ambiguity resolution scheme with 3 User IDs
        #   :user ID packet: "Test Expiration I <test@expir.one>"
        #   :user ID packet: "Test Expiration II <test@expir.two>"
        #   :user ID packet: "Test Expiration III <test@expir.three>"
        # User ID packets are ordered by their creation time in ascending order.
        # "Test Expiration II" has the primary user ID flag set and therefor has
        # the highest priority.
        key = _assign_certified_key_info(self.raw_expired_key_bundle)
        self.assertTrue(key["validity_period"] == 87901)  # ~ 1 day

        # Test ambiguity resolution scheme with 2 User IDs
        #   :user ID packet: "Test Expiration III <test@expir.three>"
        #   :user ID packet: "Test Expiration I <test@expir.one>"
        # User ID packets are ordered by their creation time in descending order.
        # Neither packet has the primary user ID flag set.
        # "Test Expiration III" has the highest priority.
        raw_key_bundle = deepcopy(self.raw_expired_key_bundle)
        user_id_items = list(
            reversed(raw_key_bundle[PACKET_TYPE_USER_ID].items())
        )
        del user_id_items[1]
        raw_key_bundle[PACKET_TYPE_USER_ID] = OrderedDict(user_id_items)
        key = _assign_certified_key_info(raw_key_bundle)
        self.assertTrue(key["validity_period"] == 87901)  # ~ 1 day

    def test_get_verified_subkeys_errors(self):
        """Test _get_verified_subkeys errors with manually crafted data based on
        real gpg key data (see self.raw_key_bundle)."""

        # Tamper with subkey (change version number) to trigger key parsing error
        bad_subkey_bundle = deepcopy(self.raw_key_bundle)
        packet, packet_data = bad_subkey_bundle[PACKET_TYPE_SUB_KEY].popitem()
        packet = bytes(
            packet[: packet_data["header_len"]]
            + bytearray([0x03])
            + packet[packet_data["header_len"] + 1 :]
        )
        bad_subkey_bundle[PACKET_TYPE_SUB_KEY][packet] = packet_data

        # Add bogus sig to trigger sig parsing error
        wrong_sig_bundle = deepcopy(self.raw_key_bundle)
        packet, packet_data = wrong_sig_bundle[PACKET_TYPE_SUB_KEY].popitem()
        # NOTE: We can't only pass the bogus sig, because that would also trigger
        # the not enough sigs error (see not_enough_sigs_bundle) and mock only
        # lets us assert for the most recent log statement
        packet_data["signatures"].append(bytearray([0b01111111, 0]))
        wrong_sig_bundle[PACKET_TYPE_SUB_KEY][packet] = packet_data

        # Remove sigs to trigger not enough sigs error
        not_enough_sigs_bundle = deepcopy(self.raw_key_bundle)
        packet, packet_data = not_enough_sigs_bundle[
            PACKET_TYPE_SUB_KEY
        ].popitem()
        packet_data["signatures"] = []
        not_enough_sigs_bundle[PACKET_TYPE_SUB_KEY][packet] = packet_data

        # Duplicate sig to trigger wrong amount signatures
        too_many_sigs_bundle = deepcopy(self.raw_key_bundle)
        packet, packet_data = too_many_sigs_bundle[
            PACKET_TYPE_SUB_KEY
        ].popitem()
        packet_data["signatures"] = packet_data["signatures"] * 2
        too_many_sigs_bundle[PACKET_TYPE_SUB_KEY][packet] = packet_data

        # Tamper with primary key to trigger signature verification error
        invalid_sig_bundle = deepcopy(self.raw_key_bundle)
        invalid_sig_bundle[PACKET_TYPE_PRIMARY_KEY][
            "packet"
        ] = invalid_sig_bundle[PACKET_TYPE_PRIMARY_KEY]["packet"][:-1]

        test_data = [
            (bad_subkey_bundle, "Pubkey packet version '3' not supported"),
            (wrong_sig_bundle, "Expected packet 2, but got 63 instead"),
            (
                not_enough_sigs_bundle,
                "wrong amount of key binding signatures (0)",
            ),
            (
                too_many_sigs_bundle,
                "wrong amount of key binding signatures (2)",
            ),
            (invalid_sig_bundle, "invalid key binding signature"),
        ]

        for bundle, expected_msg in test_data:
            with patch("securesystemslib.gpg.common.log") as mock_log:
                _get_verified_subkeys(bundle)
                msg = str(mock_log.info.call_args[0][0])
                self.assertTrue(
                    expected_msg in msg,
                    "'{}' not in '{}'".format(  # pylint: disable=consider-using-f-string
                        expected_msg, msg
                    ),
                )

    def test_get_verified_subkeys(self):
        """Test correct assignment of subkey expiration date in
        gpg.common._get_verified_subkeys using real gpg data."""
        subkeys = _get_verified_subkeys(self.raw_expired_key_bundle)
        # Test subkey with validity period 175451, i.e. ~ 2 days
        self.assertTrue(
            subkeys["0ce427fa3f0f50bc83a4a760ed95e1581691db4d"].get(
                "validity_period"
            )
            == 175451
        )

        # Test subkey  without validity period, i.e. it does not expire
        self.assertTrue(
            subkeys[  # pylint: disable=singleton-comparison
                "70cfabf1e2f1dc60ac5c7bca10cd20d3d5bcb6ef"
            ].get("validity_period")
            == None
        )

    def test_get_pubkey_bundle_errors(self):
        """Test correct error raising in get_pubkey_bundle."""
        # Call without key data
        with self.assertRaises(KeyNotFoundError):
            get_pubkey_bundle(None, "deadbeef")

        # Pass wrong keyid with valid gpg data to trigger KeyNotFoundError.
        not_associated_keyid = "8465A1E2E0FB2B40ADB2478E18FB3F537E0C8A17"
        with self.assertRaises(KeyNotFoundError):
            get_pubkey_bundle(self.raw_key_data, not_associated_keyid)

    def test_parse_signature_packet_errors(self):
        """Test parse_signature_packet errors with manually crafted data."""

        # passed data | expected error message
        test_data = [
            (
                bytearray([0b01000010, 1, 255]),
                "Signature version '255' not supported",
            ),
            (
                bytearray([0b01000010, 2, 4, 255]),
                "Signature type '255' not supported",
            ),
            (
                bytearray([0b01000010, 3, 4, 0, 255]),
                "Signature algorithm '255' not supported",
            ),
            (
                bytearray([0b01000010, 4, 4, 0, 1, 255]),
                "Hash algorithm '255' not supported",
            ),
        ]
        for data, expected_error_str in test_data:
            with self.assertRaises(ValueError) as ctx:
                parse_signature_packet(data)
            self.assertTrue(
                expected_error_str in str(ctx.exception),
                "'{}' not in '{}'".format(  # pylint: disable=consider-using-f-string
                    expected_error_str, str(ctx.exception)
                ),
            )


@unittest.skipIf(not have_gpg(), "gpg not found")
class TestGPGRSA(unittest.TestCase):
    """Test signature creation, verification and key export from the gpg
    module"""

    default_keyid = "8465A1E2E0FB2B40ADB2478E18FB3F537E0C8A17"
    signing_subkey_keyid = "C5A0ABE6EC19D0D65F85E2C39BE9DF5131D924E9"
    encryption_subkey_keyid = "6A112FD3390B2E53AFC2E57F8FC8E12099AECEEA"
    unsupported_subkey_keyid = "611A9B648E16F54E8A7FAD5DA51E8CDF3B06524F"
    expired_key_keyid = "E8AC80C924116DABB51D4B987CB07D6D2C199C7C"

    keyid_768C43 = "7B3ABB26B97B655AB9296BD15B0BD02E1C768C43"  # pylint: disable=invalid-name

    @classmethod
    def setUpClass(self):  # pylint: disable=bad-classmethod-argument
        # Create directory to run the tests without having everything blow up
        self.working_dir = os.getcwd()

        # Find demo files
        gpg_keyring_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "gpg_keyrings", "rsa"
        )

        self.test_dir = os.path.realpath(tempfile.mkdtemp())
        self.gnupg_home = os.path.join(self.test_dir, "rsa")
        shutil.copytree(gpg_keyring_path, self.gnupg_home)
        os.chdir(self.test_dir)

    @classmethod
    def tearDownClass(self):  # pylint: disable=bad-classmethod-argument
        """Change back to initial working dir and remove temp test directory."""
        os.chdir(self.working_dir)
        shutil.rmtree(
            self.test_dir, onerror=GPGTestUtils.ignore_not_found_error
        )

    def test_export_pubkey_error(self):
        """Test correct error is raised if function called incorrectly."""
        with self.assertRaises(ValueError):
            export_pubkey("not-a-key-id")

    def test_export_pubkey(self):
        """export a public key and make sure the parameters are the right ones:

        since there's very little we can do to check rsa key parameters are right
        we pre-exported the public key to an ssh key, which we can load with
        cryptography for the sake of comparison"""

        # export our gpg key, using our functions
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)
        our_exported_key = rsa_create_pubkey(key_data)

        # load the equivalent ssh key, and make sure that we get the same RSA key
        # parameters
        ssh_key_basename = (
            "{}.ssh".format(  # pylint: disable=consider-using-f-string
                self.default_keyid
            )
        )
        ssh_key_path = os.path.join(self.gnupg_home, ssh_key_basename)
        with open(ssh_key_path, "rb") as fp:
            keydata = fp.read()

        ssh_key = serialization.load_ssh_public_key(
            keydata, backends.default_backend()
        )

        self.assertEqual(
            ssh_key.public_numbers().n, our_exported_key.public_numbers().n
        )
        self.assertEqual(
            ssh_key.public_numbers().e, our_exported_key.public_numbers().e
        )

        subkey_keyids = list(key_data["subkeys"].keys())
        # We export the whole master key bundle which must contain the subkeys
        self.assertTrue(self.signing_subkey_keyid.lower() in subkey_keyids)
        # Currently we do not exclude encryption subkeys
        self.assertTrue(self.encryption_subkey_keyid.lower() in subkey_keyids)
        # However we do exclude subkeys, whose algorithm we do not support
        self.assertFalse(self.unsupported_subkey_keyid.lower() in subkey_keyids)

        # When passing the subkey keyid we also export the whole keybundle
        key_data2 = export_pubkey(
            self.signing_subkey_keyid, homedir=self.gnupg_home
        )
        self.assertDictEqual(key_data, key_data2)

    def test_export_pubkeys(self):
        """Test export multiple pubkeys at once."""
        key_dict = export_pubkeys(
            [self.default_keyid, self.keyid_768C43], homedir=self.gnupg_home
        )

        ANY_PUBKEY_DICT_SCHEMA.check_match(key_dict)
        self.assertListEqual(
            sorted([self.default_keyid.lower(), self.keyid_768C43.lower()]),
            sorted(key_dict.keys()),
        )

    def test_gpg_sign_and_verify_object_with_default_key(self):
        """Create a signature using the default key on the keyring"""

        test_data = b"test_data"
        wrong_data = b"something malicious"

        signature = create_signature(test_data, homedir=self.gnupg_home)
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)

        self.assertTrue(verify_signature(signature, key_data, test_data))
        self.assertFalse(verify_signature(signature, key_data, wrong_data))

    def test_gpg_sign_and_verify_object(self):
        """Create a signature using a specific key on the keyring"""

        test_data = b"test_data"
        wrong_data = b"something malicious"

        signature = create_signature(
            test_data, keyid=self.default_keyid, homedir=self.gnupg_home
        )
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)
        self.assertTrue(verify_signature(signature, key_data, test_data))
        self.assertFalse(verify_signature(signature, key_data, wrong_data))

    def test_gpg_sign_and_verify_object_default_keyring(self):
        """Sign/verify using keyring from envvar."""

        test_data = b"test_data"

        gnupg_home_backup = os.environ.get("GNUPGHOME")
        os.environ["GNUPGHOME"] = self.gnupg_home

        signature = create_signature(test_data, keyid=self.default_keyid)
        key_data = export_pubkey(self.default_keyid)
        self.assertTrue(verify_signature(signature, key_data, test_data))

        # Reset GNUPGHOME
        if gnupg_home_backup:
            os.environ["GNUPGHOME"] = gnupg_home_backup
        else:
            del os.environ["GNUPGHOME"]

    def test_create_signature_with_expired_key(self):
        """Test signing with expired key raises gpg CommandError."""
        with self.assertRaises(CommandError) as ctx:
            create_signature(
                b"livestock",
                keyid=self.expired_key_keyid,
                homedir=self.gnupg_home,
            )

        expected = "returned non-zero exit status '2'"
        self.assertTrue(
            expected in str(ctx.exception),
            "{} not in {}".format(  # pylint: disable=consider-using-f-string
                expected, ctx.exception
            ),
        )

    def test_verify_signature_with_expired_key(self):
        """Test sig verification with expired key raises KeyExpirationError."""
        signature = {
            "keyid": self.expired_key_keyid,
            "other_headers": "deadbeef",
            "signature": "deadbeef",
        }
        content = b"livestock"
        key = export_pubkey(self.expired_key_keyid, homedir=self.gnupg_home)

        with self.assertRaises(KeyExpirationError) as ctx:
            verify_signature(signature, key, content)

        expected = (
            "GPG key 'e8ac80c924116dabb51d4b987cb07d6d2c199c7c' "
            "created on '2019-03-25 12:46 UTC' with validity period '1 day, "
            "0:25:01' expired on '2019-03-26 13:11 UTC'."
        )
        self.assertTrue(
            expected == str(ctx.exception),
            "\nexpected: {}"  # pylint: disable=consider-using-f-string
            "\ngot:      {}".format(  # pylint: disable=consider-using-f-string
                expected, ctx.exception
            ),
        )


@unittest.skipIf(not have_gpg(), "gpg not found")
class TestGPGDSA(unittest.TestCase):
    """Test signature creation, verification and key export from the gpg
    module"""

    default_keyid = "C242A830DAAF1C2BEF604A9EF033A3A3E267B3B1"

    @classmethod
    def setUpClass(self):  # pylint: disable=bad-classmethod-argument
        # Create directory to run the tests without having everything blow up
        self.working_dir = os.getcwd()
        self.test_dir = os.path.realpath(tempfile.mkdtemp())
        self.gnupg_home = os.path.join(self.test_dir, "dsa")

        # Find keyrings
        keyrings = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "gpg_keyrings", "dsa"
        )

        shutil.copytree(keyrings, self.gnupg_home)
        os.chdir(self.test_dir)

    @classmethod
    def tearDownClass(self):  # pylint: disable=bad-classmethod-argument
        """Change back to initial working dir and remove temp test directory."""
        os.chdir(self.working_dir)
        shutil.rmtree(
            self.test_dir, onerror=GPGTestUtils.ignore_not_found_error
        )

    def test_export_pubkey(self):
        """export a public key and make sure the parameters are the right ones:

        since there's very little we can do to check key parameters are right
        we pre-exported the public key to an x.509 SubjectPublicKeyInfo key,
        which we can load with cryptography for the sake of comparison"""

        # export our gpg key, using our functions
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)
        our_exported_key = dsa_create_pubkey(key_data)

        # load same key, pre-exported with 3rd-party tooling
        pem_key_basename = (
            "{}.pem".format(  # pylint: disable=consider-using-f-string
                self.default_keyid
            )
        )
        pem_key_path = os.path.join(self.gnupg_home, pem_key_basename)
        with open(pem_key_path, "rb") as fp:
            keydata = fp.read()

        pem_key = serialization.load_pem_public_key(
            keydata, backends.default_backend()
        )

        # make sure keys match
        self.assertEqual(
            pem_key.public_numbers().y, our_exported_key.public_numbers().y
        )
        self.assertEqual(
            pem_key.public_numbers().parameter_numbers.g,
            our_exported_key.public_numbers().parameter_numbers.g,
        )
        self.assertEqual(
            pem_key.public_numbers().parameter_numbers.q,
            our_exported_key.public_numbers().parameter_numbers.q,
        )
        self.assertEqual(
            pem_key.public_numbers().parameter_numbers.p,
            our_exported_key.public_numbers().parameter_numbers.p,
        )

    def test_gpg_sign_and_verify_object_with_default_key(self):
        """Create a signature using the default key on the keyring"""

        test_data = b"test_data"
        wrong_data = b"something malicious"

        signature = create_signature(test_data, homedir=self.gnupg_home)
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)

        self.assertTrue(verify_signature(signature, key_data, test_data))
        self.assertFalse(verify_signature(signature, key_data, wrong_data))

    def test_gpg_sign_and_verify_object(self):
        """Create a signature using a specific key on the keyring"""

        test_data = b"test_data"
        wrong_data = b"something malicious"

        signature = create_signature(
            test_data, keyid=self.default_keyid, homedir=self.gnupg_home
        )
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)

        self.assertTrue(verify_signature(signature, key_data, test_data))
        self.assertFalse(verify_signature(signature, key_data, wrong_data))


@unittest.skipIf(not have_gpg(), "gpg not found")
class TestGPGEdDSA(unittest.TestCase):
    """Test signature creation, verification and key export from the gpg
    module"""

    default_keyid = "4E630F84838BF6F7447B830B22692F5FEA9E2DD2"

    @classmethod
    def setUpClass(self):  # pylint: disable=bad-classmethod-argument
        # Create directory to run the tests without having everything blow up
        self.working_dir = os.getcwd()
        self.test_dir = os.path.realpath(tempfile.mkdtemp())
        self.gnupg_home = os.path.join(self.test_dir, "dsa")

        # Find keyrings
        keyrings = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "gpg_keyrings", "eddsa"
        )

        shutil.copytree(keyrings, self.gnupg_home)
        os.chdir(self.test_dir)

    @classmethod
    def tearDownClass(self):  # pylint: disable=bad-classmethod-argument
        """Change back to initial working dir and remove temp test directory."""
        os.chdir(self.working_dir)
        shutil.rmtree(
            self.test_dir, onerror=GPGTestUtils.ignore_not_found_error
        )

    def test_gpg_sign_and_verify_object_with_default_key(self):
        """Create a signature using the default key on the keyring"""

        test_data = b"test_data"
        wrong_data = b"something malicious"

        signature = create_signature(test_data, homedir=self.gnupg_home)
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)

        self.assertTrue(verify_signature(signature, key_data, test_data))
        self.assertFalse(verify_signature(signature, key_data, wrong_data))

    def test_gpg_sign_and_verify_object_with_specific_key(self):
        """Create a signature using a specific key on the keyring"""

        test_data = b"test_data"
        wrong_data = b"something malicious"

        signature = create_signature(
            test_data, keyid=self.default_keyid, homedir=self.gnupg_home
        )
        key_data = export_pubkey(self.default_keyid, homedir=self.gnupg_home)

        self.assertTrue(verify_signature(signature, key_data, test_data))
        self.assertFalse(verify_signature(signature, key_data, wrong_data))

    def test_verify_short_signature(self):
        """Correctly verify a special-crafted short signature."""

        test_data = b"hello"
        signature_path = os.path.join(self.gnupg_home, "short.sig")

        # Read special-crafted raw gpg signature that is one byte too short
        with open(signature_path, "rb") as f:
            signature_data = f.read()

        # Check that the signature is padded upon parsing
        # NOTE: The returned signature is a hex string and thus twice as long
        signature = parse_signature_packet(signature_data)
        self.assertTrue(len(signature["signature"]) == (ED25519_SIG_LENGTH * 2))

        # Check that the signature can be successfully verified
        key = export_pubkey(self.default_keyid, homedir=self.gnupg_home)
        self.assertTrue(verify_signature(signature, key, test_data))


class TestVersion(unittest.TestCase):
    """Tests for the Version utility class."""

    def test_version_roundtrip_string(self):
        """Version parses and formats strings correctly."""
        for value, expected in [
            ("1.3.0", Version(1, 3, 0)),
            ("1.3.1", Version(1, 3, 1)),
            ("1.3.22", Version(1, 3, 22)),
        ]:
            self.assertEqual(Version.from_string(value), expected)
            self.assertEqual(str(expected), value)

    def test_version_from_string_invalid(self):
        """Version.from_string rejects invalid inputs."""
        for value in [
            "1.3",
            "1.33.0",
            "1.3.-1",
            "1.3.1a",
        ]:
            with self.assertRaises(
                ValueError, msg=f"expected error for input '{value}'"
            ):
                Version.from_string(value)


if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""
<Program Name>
  test_util.py

<Author>
  Konstantin Andrianov.

<Started>
  February 1, 2013.

<Copyright>
  See LICENSE for licensing information.

<Purpose>
  Unit test for 'util.py'
"""

import logging
import os
import stat
import tempfile
import timeit
import unittest

import securesystemslib.hash
import securesystemslib.settings
import securesystemslib.util
from securesystemslib import exceptions, unittest_toolbox

logger = logging.getLogger(__name__)


class TestUtil(
    unittest_toolbox.Modified_TestCase
):  # pylint: disable=missing-class-docstring
    def setUp(self):
        unittest_toolbox.Modified_TestCase.setUp(self)
        self.temp_fileobj = tempfile.TemporaryFile()

    def tearDown(self):
        unittest_toolbox.Modified_TestCase.tearDown(self)
        self.temp_fileobj.close()

    def test_B1_get_file_details(self):
        # Goal: Verify proper output given certain expected/unexpected input.

        # Making a temporary file.
        filepath = self.make_temp_data_file()

        # Computing the hash and length of the tempfile.
        digest_object = securesystemslib.hash.digest_filename(
            filepath, algorithm="sha256"
        )
        file_hash = {"sha256": digest_object.hexdigest()}
        file_length = os.path.getsize(filepath)

        # Test: Expected input.
        self.assertEqual(
            securesystemslib.util.get_file_details(filepath),
            (file_length, file_hash),
        )

        # Test: Incorrect input.
        bogus_inputs = [
            self.random_string(),
            1234,
            [self.random_string()],
            {"a": "b"},
            None,
        ]

        for bogus_input in bogus_inputs:
            if isinstance(bogus_input, str):
                self.assertRaises(
                    securesystemslib.exceptions.StorageError,
                    securesystemslib.util.get_file_details,
                    bogus_input,
                )
            else:
                self.assertRaises(
                    securesystemslib.exceptions.FormatError,
                    securesystemslib.util.get_file_details,
                    bogus_input,
                )

    def test_B2_get_file_hashes(self):
        # Goal: Verify proper output given certain expected/unexpected input.

        # Making a temporary file.
        filepath = self.make_temp_data_file()

        # Computing the hash of the tempfile.
        digest_object = securesystemslib.hash.digest_filename(
            filepath, algorithm="sha256"
        )
        file_hash = {"sha256": digest_object.hexdigest()}

        # Test: Expected input.
        self.assertEqual(
            securesystemslib.util.get_file_hashes(filepath), file_hash
        )

        # Test: Incorrect input.
        bogus_inputs = [
            self.random_string(),
            1234,
            [self.random_string()],
            {"a": "b"},
            None,
        ]

        for bogus_input in bogus_inputs:
            if isinstance(bogus_input, str):
                self.assertRaises(
                    securesystemslib.exceptions.StorageError,
                    securesystemslib.util.get_file_hashes,
                    bogus_input,
                )
            else:
                self.assertRaises(
                    securesystemslib.exceptions.FormatError,
                    securesystemslib.util.get_file_hashes,
                    bogus_input,
                )

    def test_B3_get_file_length(self):
        # Goal: Verify proper output given certain expected/unexpected input.

        # Making a temporary file.
        filepath = self.make_temp_data_file()

        # Computing the length of the tempfile.
        file_length = os.path.getsize(filepath)

        # Test: Expected input.
        self.assertEqual(
            securesystemslib.util.get_file_length(filepath), file_length
        )

        # Test: Incorrect input.
        bogus_inputs = [
            self.random_string(),
            1234,
            [self.random_string()],
            {"a": "b"},
            None,
        ]

        for bogus_input in bogus_inputs:
            if isinstance(bogus_input, str):
                self.assertRaises(
                    securesystemslib.exceptions.StorageError,
                    securesystemslib.util.get_file_length,
                    bogus_input,
                )
            else:
                self.assertRaises(
                    securesystemslib.exceptions.FormatError,
                    securesystemslib.util.get_file_length,
                    bogus_input,
                )

    def test_B4_ensure_parent_dir(self):
        existing_parent_dir = self.make_temp_directory()
        non_existing_parent_dir = os.path.join(existing_parent_dir, "a", "b")

        for parent_dir in [
            existing_parent_dir,
            non_existing_parent_dir,
            12,
            [3],
        ]:
            if isinstance(parent_dir, str):
                securesystemslib.util.ensure_parent_dir(
                    os.path.join(parent_dir, "a.txt")
                )
                self.assertTrue(os.path.isdir(parent_dir))

            else:
                self.assertRaises(
                    securesystemslib.exceptions.FormatError,
                    securesystemslib.util.ensure_parent_dir,
                    parent_dir,
                )

        # Check that when a folder cannot be created a StorageError is thrown
        with self.assertRaises(securesystemslib.exceptions.StorageError):
            securesystemslib.util.ensure_parent_dir("/a/file.txt")

        # When we call ensure_parent_dir with filepath arg like "a.txt",
        # then the directory of that filepath will be an empty string.
        # We want to make sure that securesyslib.storage.create_folder()
        # won't be called with an empty string and thus raise an exception.
        # If an exception is thrown the test will fail.
        securesystemslib.util.ensure_parent_dir("a.txt")

    def test_B5_file_in_confined_directories(self):
        # Goal: Provide invalid input for 'filepath' and 'confined_directories'.
        # Include inputs like: '[1, 2, "a"]' and such...
        # Reference to 'file_in_confined_directories()' to improve readability.
        in_confined_directory = (
            securesystemslib.util.file_in_confined_directories
        )
        list_of_confined_directories = ["a", 12, {"a": "a"}, [1]]
        list_of_filepaths = [12, ["a"], {"a": "a"}, "a"]
        for bogus_confined_directory in list_of_confined_directories:
            for filepath in list_of_filepaths:
                self.assertRaises(
                    securesystemslib.exceptions.FormatError,
                    in_confined_directory,
                    filepath,
                    bogus_confined_directory,
                )

        # Test: Inputs that evaluate to False.
        confined_directories = ["a/b/", "a/b/c/d/"]
        self.assertFalse(
            in_confined_directory("a/b/c/1.txt", confined_directories)
        )

        confined_directories = ["a/b/c/d/e/"]
        self.assertFalse(in_confined_directory("a", confined_directories))
        self.assertFalse(in_confined_directory("a/b", confined_directories))
        self.assertFalse(in_confined_directory("a/b/c", confined_directories))
        self.assertFalse(in_confined_directory("a/b/c/d", confined_directories))
        # Below, 'e' is a file in the 'a/b/c/d/' directory.
        self.assertFalse(
            in_confined_directory("a/b/c/d/e", confined_directories)
        )

        # Test: Inputs that evaluate to True.
        self.assertTrue(in_confined_directory("a/b/c.txt", [""]))
        self.assertTrue(in_confined_directory("a/b/c.txt", ["a/b/"]))
        self.assertTrue(in_confined_directory("a/b/c.txt", ["x", ""]))
        self.assertTrue(in_confined_directory("a/b/c/..", ["a/"]))

    def test_B7_load_json_string(self):
        # Test normal case.
        data = ["a", {"b": ["c", None, 30.3, 29]}]
        json_string = securesystemslib.util.json.dumps(data)
        self.assertEqual(
            data, securesystemslib.util.load_json_string(json_string)
        )

        # Test invalid arguments.
        self.assertRaises(
            securesystemslib.exceptions.Error,
            securesystemslib.util.load_json_string,
            8,
        )
        invalid_json_string = json_string + "."
        self.assertRaises(
            securesystemslib.exceptions.Error,
            securesystemslib.util.load_json_string,
            invalid_json_string,
        )

    def test_B8_load_json_file(self):
        data = ["a", {"b": ["c", None, 30.3, 29]}]
        filepath = self.make_temp_file()
        fileobj = (
            open(  # pylint: disable=unspecified-encoding,consider-using-with
                filepath, "wt"
            )
        )
        securesystemslib.util.json.dump(data, fileobj)
        fileobj.close()
        self.assertEqual(data, securesystemslib.util.load_json_file(filepath))

        # Improperly formatted arguments.
        for bogus_arg in [1, [b"a"], {"a": b"b"}]:
            self.assertRaises(
                securesystemslib.exceptions.FormatError,
                securesystemslib.util.load_json_file,
                bogus_arg,
            )

        # Non-existent path.
        self.assertRaises(
            securesystemslib.exceptions.StorageError,
            securesystemslib.util.load_json_file,
            "non-existent.json",
        )

        # Invalid JSON content.
        filepath_bad_data = self.make_temp_file()
        fileobj = (
            open(  # pylint: disable=unspecified-encoding,consider-using-with
                filepath_bad_data, "wt"
            )
        )
        fileobj.write("junk data")
        fileobj.close()

        self.assertRaises(
            securesystemslib.exceptions.Error,
            securesystemslib.util.load_json_file,
            filepath_bad_data,
        )

    def test_B9_persist_temp_file(self):
        # Destination directory to save the temporary file in.
        dest_temp_dir = self.make_temp_directory()

        # Test the default of persisting the file and closing the tmpfile
        dest_path = os.path.join(dest_temp_dir, self.random_string())
        tmpfile = tempfile.TemporaryFile()
        tmpfile.write(self.random_string().encode("utf-8"))

        # Write a file with restricted permissions
        securesystemslib.util.persist_temp_file(
            tmpfile, dest_path, restrict=True
        )
        self.assertTrue(dest_path)

        # Need to set also the stat.S_IFREG bit to match the st_mode output
        # stat.S_IFREG - Regular file
        expected_mode = stat.S_IFREG | stat.S_IRUSR | stat.S_IWUSR
        if os.name == "nt":
            # Windows only supports setting the read-only attribute.
            expected_mode = (
                stat.S_IFREG
                | stat.S_IWUSR
                | stat.S_IRUSR
                | stat.S_IWGRP
                | stat.S_IRGRP
                | stat.S_IWOTH
                | stat.S_IROTH
            )
        self.assertEqual(os.stat(dest_path).st_mode, expected_mode)
        self.assertTrue(tmpfile.closed)

        # Test persisting a file without automatically closing the tmpfile
        dest_path2 = os.path.join(dest_temp_dir, self.random_string())
        tmpfile = tempfile.TemporaryFile()
        tmpfile.write(self.random_string().encode("utf-8"))
        securesystemslib.util.persist_temp_file(
            tmpfile, dest_path2, should_close=False
        )
        self.assertFalse(tmpfile.closed)

        # Test persisting a file with an empty filename
        with self.assertRaises(exceptions.StorageError):
            securesystemslib.util.persist_temp_file(tmpfile, "")

        tmpfile.close()

    def test_C5_unittest_toolbox_make_temp_directory(self):
        # Verify that the tearDown function does not fail when
        # unittest_toolbox.make_temp_directory deletes the generated temp directory
        # here.
        temp_directory = self.make_temp_directory()
        os.rmdir(temp_directory)

    def test_c5_unittest_toolbox_random_path(self):
        # Verify that a random path can be generated with unittest_toolbox.
        random_path = self.random_path(length=10)
        self.assertTrue(
            securesystemslib.formats.PATH_SCHEMA.matches(random_path)
        )
        self.assertTrue(  # pylint: disable=redundant-unittest-assert
            10, len(random_path)
        )

    def test_digests_are_equal(self):
        digest = (
            "e3b0c44298fc1c149afbf4c8996fb92427ae41e4649b934ca495991b7852b855"
        )

        # Normal case: test for digests that are equal.
        self.assertTrue(securesystemslib.util.digests_are_equal(digest, digest))

        # Normal case: test for digests that are unequal.
        self.assertFalse(
            securesystemslib.util.digests_are_equal(digest, "0a8df1")
        )

        # Test for invalid arguments.
        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.util.digests_are_equal,
            7,
            digest,
        )
        self.assertRaises(
            securesystemslib.exceptions.FormatError,
            securesystemslib.util.digests_are_equal,
            digest,
            7,
        )

        # Test that digests_are_equal() takes the same amount of time to compare
        # equal and unequal arguments.
        runtime = timeit.timeit(
            'digests_are_equal("ab8df", "ab8df")',
            setup="from securesystemslib.util import digests_are_equal",
            number=100000,
        )

        runtime2 = timeit.timeit(
            'digests_are_equal("ab8df", "1b8df")',
            setup="from securesystemslib.util import digests_are_equal",
            number=100000,
        )

        runtime3 = timeit.timeit('"ab8df" == "ab8df"', number=100000)

        runtime4 = timeit.timeit('"ab8df" == "1b8df"', number=1000000)

        # The ratio for the 'digest_are_equal' runtimes should be at or near 1.
        ratio_digests_are_equal = abs(runtime2 / runtime)

        # The ratio for the variable-time runtimes should be (>1) & at or near 10?
        ratio_variable_compare = abs(runtime4 / runtime3)

        self.assertTrue(ratio_digests_are_equal < ratio_variable_compare)


# Run unit test.
if __name__ == "__main__":
    unittest.main()



#!/usr/bin/env python

"""Test cases for "signer.py". """

import copy
import os
import shutil
import tempfile
import unittest
from typing import Any, Dict, Optional

import securesystemslib.keys as KEYS
from securesystemslib.exceptions import (
    CryptoError,
    FormatError,
    UnsupportedAlgorithmError,
    UnverifiedSignatureError,
)
from securesystemslib.gpg.constants import have_gpg
from securesystemslib.gpg.functions import export_pubkey
from securesystemslib.gpg.functions import verify_signature as verify_sig
from securesystemslib.signer import (
    KEY_FOR_TYPE_AND_SCHEME,
    SIGNER_FOR_URI_SCHEME,
    GPGSigner,
    Key,
    SecretsHandler,
    Signature,
    Signer,
    SSlibKey,
    SSlibSigner,
)


class TestKey(unittest.TestCase):
    """Key tests. See many more tests in python-tuf test suite"""

    def test_key_from_to_dict(self):
        """Test to/from_dict for known keytype/scheme combos"""
        for (keytype, scheme), key_impl in KEY_FOR_TYPE_AND_SCHEME.items():
            keydict = {
                "keytype": keytype,
                "scheme": scheme,
                "extra": "somedata",
                "keyval": {
                    "public": "pubkeyval",
                    "foo": "bar",
                },
            }

            key = Key.from_dict("aa", copy.deepcopy(keydict))
            self.assertIsInstance(key, key_impl)
            self.assertDictEqual(keydict, key.to_dict())

    def test_sslib_key_from_dict_invalid(self):
        """Test from_dict for invalid data"""
        invalid_dicts = [
            {"scheme": "ed25519", "keyval": {"public": "abc"}},
            {"keytype": "ed25519", "keyval": {"public": "abc"}},
            {"keytype": "ed25519", "scheme": "ed25519"},
            {"keytype": "ed25519", "scheme": "ed25519", "keyval": {"x": "y"}},
            {
                "keytype": "ed25519",
                "scheme": "ed25519",
                "keyval": {"public": b"abc"},
            },
        ]
        for keydict in invalid_dicts:
            with self.assertRaises((KeyError, ValueError)):
                Key.from_dict("aa", keydict)

    def test_key_verify_signature(self):
        sigdict = {
            "keyid": "e33221e745d40465d1efc0215d6db83e5fdb83ea16e1fb894d09d6d96c456f3b",
            "sig": "3fc91f5411a567d6a7f28b7fbb9ba6d60b1e2a1b64d8af0b119650015d86bb5a55e57c0e2c995a9b4a332b8f435703e934c0e6ce69fe6674a8ce68719394a40b",
        }
        keydict = {
            "keytype": "ed25519",
            "scheme": "ed25519",
            "keyval": {
                "public": "8ae43d22b8e0fbf4a48fa3490d31b4d389114f5dc1039c918f075427f4100759",
            },
        }
        key = Key.from_dict(
            "e33221e745d40465d1efc0215d6db83e5fdb83ea16e1fb894d09d6d96c456f3b",
            keydict,
        )
        sig = Signature.from_dict(sigdict)

        key.verify_signature(sig, b"DATA")
        with self.assertRaises(UnverifiedSignatureError):
            key.verify_signature(sig, b"NOT DATA")

    def test_unsupported_key(self):
        keydict = {
            "keytype": "custom",
            "scheme": "ed25519",
            "keyval": {
                "public": "8ae43d22b8e0fbf4a48fa3490d31b4d389114f5dc1039c918f075427f4100759",
            },
        }
        with self.assertRaises(ValueError):
            Key.from_dict(
                "e33221e745d40465d1efc0215d6db83e5fdb83ea16e1fb894d09d6d96c456f3b",
                keydict,
            )

    def test_custom_key(self):
        class CustomKey(SSlibKey):
            """Fake keytype that actually uses ed25519 under the hood"""

            @classmethod
            def from_dict(
                cls, keyid: str, key_dict: Dict[str, Any]
            ) -> "CustomKey":
                assert key_dict.pop("keytype") == "custom"
                keytype = "ed25519"
                scheme = key_dict.pop("scheme")
                keyval = key_dict.pop("keyval")
                return cls(keyid, keytype, scheme, keyval, key_dict)

            def to_dict(self) -> Dict[str, Any]:
                return {
                    "keytype": "custom",
                    "scheme": self.scheme,
                    "keyval": self.keyval,
                    **self.unrecognized_fields,
                }

        # register custom key type
        KEY_FOR_TYPE_AND_SCHEME[("custom", "ed25519")] = CustomKey

        # setup
        sig = Signature.from_dict(
            {
                "keyid": "e33221e745d40465d1efc0215d6db83e5fdb83ea16e1fb894d09d6d96c456f3b",
                "sig": "3fc91f5411a567d6a7f28b7fbb9ba6d60b1e2a1b64d8af0b119650015d86bb5a55e57c0e2c995a9b4a332b8f435703e934c0e6ce69fe6674a8ce68719394a40b",
            }
        )

        keydict = {
            "keytype": "custom",
            "scheme": "ed25519",
            "keyval": {
                "public": "8ae43d22b8e0fbf4a48fa3490d31b4d389114f5dc1039c918f075427f4100759",
            },
        }
        key = Key.from_dict(
            "e33221e745d40465d1efc0215d6db83e5fdb83ea16e1fb894d09d6d96c456f3b",
            keydict,
        )

        # test that CustomKey is used and that it works
        self.assertIsInstance(key, CustomKey)
        key.verify_signature(sig, b"DATA")
        with self.assertRaises(UnverifiedSignatureError):
            key.verify_signature(sig, b"NOT DATA")

        del KEY_FOR_TYPE_AND_SCHEME[("custom", "ed25519")]


class TestSigner(unittest.TestCase):
    """Test Signer and SSlibSigner functionality"""

    @classmethod
    def setUpClass(cls):
        cls.keys = [
            KEYS.generate_rsa_key(),
            KEYS.generate_ed25519_key(),
            KEYS.generate_ecdsa_key(),
            KEYS.generate_sphincs_key(),
        ]
        cls.DATA = b"DATA"

        # pylint: disable=consider-using-with
        cls.testdir = tempfile.TemporaryDirectory()

    @classmethod
    def tearDownClass(cls):
        cls.testdir.cleanup()

    def test_signer_sign_with_incorrect_uri(self):
        pubkey = SSlibKey.from_securesystemslib_key(self.keys[0])
        with self.assertRaises(ValueError):
            # unknown uri
            Signer.from_priv_key_uri("unknownscheme:x", pubkey)

        with self.assertRaises(ValueError):
            # env variable not defined
            Signer.from_priv_key_uri("envvar:NONEXISTENTVAR", pubkey)

        with self.assertRaises(ValueError):
            # no "encrypted" param
            Signer.from_priv_key_uri("file:path/to/privkey", pubkey)

        with self.assertRaises(OSError):
            # file not found
            uri = "file:nonexistentfile?encrypted=false"
            Signer.from_priv_key_uri(uri, pubkey)

    def test_signer_sign_with_envvar_uri(self):
        for key in self.keys:
            # setup
            pubkey = SSlibKey.from_securesystemslib_key(key)
            os.environ["PRIVKEY"] = key["keyval"]["private"]

            # test signing
            signer = Signer.from_priv_key_uri("envvar:PRIVKEY", pubkey)
            sig = signer.sign(self.DATA)

            pubkey.verify_signature(sig, self.DATA)
            with self.assertRaises(UnverifiedSignatureError):
                pubkey.verify_signature(sig, b"NOT DATA")

    def test_signer_sign_with_file_uri(self):
        for key in self.keys:
            # setup
            pubkey = SSlibKey.from_securesystemslib_key(key)
            # let teardownclass handle the file removal
            with tempfile.NamedTemporaryFile(
                dir=self.testdir.name, delete=False
            ) as f:
                f.write(key["keyval"]["private"].encode())

            # test signing with unencrypted key
            uri = f"file:{f.name}?encrypted=false"
            signer = Signer.from_priv_key_uri(uri, pubkey)
            sig = signer.sign(self.DATA)

            pubkey.verify_signature(sig, self.DATA)
            with self.assertRaises(UnverifiedSignatureError):
                pubkey.verify_signature(sig, b"NOT DATA")

    def test_signer_sign_with_enc_file_uri(self):
        for key in self.keys:
            # setup
            pubkey = SSlibKey.from_securesystemslib_key(key)
            privkey = KEYS.encrypt_key(key, "hunter2")
            # let teardownclass handle the file removal
            with tempfile.NamedTemporaryFile(
                dir=self.testdir.name, delete=False
            ) as f:
                f.write(privkey.encode())

            # test signing with encrypted key
            def secrets_handler(secret: str) -> str:
                if secret != "passphrase":
                    raise ValueError("Only prepared to return a passphrase")
                return "hunter2"

            uri = f"file:{f.name}?encrypted=true"

            signer = Signer.from_priv_key_uri(uri, pubkey, secrets_handler)
            sig = signer.sign(self.DATA)

            pubkey.verify_signature(sig, self.DATA)
            with self.assertRaises(UnverifiedSignatureError):
                pubkey.verify_signature(sig, b"NOT DATA")

            # test wrong passphrase
            def fake_handler(_) -> str:
                return "12345"

            with self.assertRaises(CryptoError):
                signer = Signer.from_priv_key_uri(uri, pubkey, fake_handler)

    def test_sslib_signer_sign(self):
        for scheme_dict in self.keys:
            # Test generation of signatures.
            sslib_signer = SSlibSigner(scheme_dict)
            sig_obj = sslib_signer.sign(self.DATA)

            # Verify signature
            verified = KEYS.verify_signature(
                scheme_dict, sig_obj.to_dict(), self.DATA
            )
            self.assertTrue(verified, "Incorrect signature.")

            # Removing private key from "scheme_dict".
            private = scheme_dict["keyval"]["private"]
            scheme_dict["keyval"]["private"] = ""
            sslib_signer.key_dict = scheme_dict

            with self.assertRaises((ValueError, FormatError)):
                sslib_signer.sign(self.DATA)

            scheme_dict["keyval"]["private"] = private

            # Test for invalid signature scheme.
            valid_scheme = scheme_dict["scheme"]
            scheme_dict["scheme"] = "invalid_scheme"
            sslib_signer = SSlibSigner(scheme_dict)

            with self.assertRaises((UnsupportedAlgorithmError, FormatError)):
                sslib_signer.sign(self.DATA)

            scheme_dict["scheme"] = valid_scheme

    def test_custom_signer(self):
        # setup
        key = self.keys[0]
        pubkey = SSlibKey.from_securesystemslib_key(key)

        class CustomSigner(SSlibSigner):
            """Custom signer with a hard coded key"""

            CUSTOM_SCHEME = "custom"

            @classmethod
            def from_priv_key_uri(
                cls,
                priv_key_uri: str,
                public_key: Key,
                secrets_handler: Optional[SecretsHandler] = None,
            ) -> "CustomSigner":
                return cls(key)

        # register custom signer
        SIGNER_FOR_URI_SCHEME[CustomSigner.CUSTOM_SCHEME] = CustomSigner

        # test signing
        signer = Signer.from_priv_key_uri("custom:foo", pubkey)
        self.assertIsInstance(signer, CustomSigner)
        sig = signer.sign(self.DATA)

        pubkey.verify_signature(sig, self.DATA)
        with self.assertRaises(UnverifiedSignatureError):
            pubkey.verify_signature(sig, b"NOT DATA")

    def test_signature_from_to_dict(self):
        signature_dict = {
            "sig": "30460221009342e4566528fcecf6a7a5d53ebacdb1df151e242f55f8775883469cb01dbc6602210086b426cc826709acfa2c3f9214610cb0a832db94bbd266fd7c5939a48064a851",
            "keyid": "11fa391a0ed7a447cbfeb4b2667e286fc248f64d5e6d0eeed2e5e23f97f9f714",
            "foo": "bar",  # unrecognized_field
        }
        sig_obj = Signature.from_dict(copy.copy(signature_dict))

        # Verify that unrecognized fields are stored correctly.
        self.assertEqual(sig_obj.unrecognized_fields, {"foo": "bar"})

        self.assertDictEqual(signature_dict, sig_obj.to_dict())

    def test_signature_eq_(self):
        signature_dict = {
            "sig": "30460221009342e4566528fcecf6a7a5d53ebacdb1df151e242f55f8775883469cb01dbc6602210086b426cc826709acfa2c3f9214610cb0a832db94bbd266fd7c5939a48064a851",
            "keyid": "11fa391a0ed7a447cbfeb4b2667e286fc248f64d5e6d0eeed2e5e23f97f9f714",
        }
        sig_obj = Signature.from_dict(signature_dict)
        sig_obj_2 = copy.deepcopy(sig_obj)

        self.assertEqual(sig_obj, sig_obj_2)

        # Assert that changing the keyid will make the objects not equal.
        sig_obj_2.keyid = None
        self.assertNotEqual(sig_obj, sig_obj_2)
        sig_obj_2.keyid = sig_obj.keyid

        # Assert that changing the signature will make the objects not equal.
        sig_obj_2.signature = None
        self.assertNotEqual(sig_obj, sig_obj_2)

        # Assert that making sig_obj_2 None will make the objects not equal.
        sig_obj_2 = None
        self.assertNotEqual(sig_obj, sig_obj_2)


@unittest.skipIf(not have_gpg(), "gpg not found")
class TestGPGRSA(unittest.TestCase):
    """Test RSA gpg signature creation and verification."""

    @classmethod
    def setUpClass(cls):
        cls.default_keyid = "8465A1E2E0FB2B40ADB2478E18FB3F537E0C8A17"
        cls.signing_subkey_keyid = "C5A0ABE6EC19D0D65F85E2C39BE9DF5131D924E9"

        # Create directory to run the tests without having everything blow up.
        cls.working_dir = os.getcwd()
        cls.test_data = b"test_data"
        cls.wrong_data = b"something malicious"

        # Find demo files.
        gpg_keyring_path = os.path.join(
            os.path.dirname(os.path.realpath(__file__)), "gpg_keyrings", "rsa"
        )

        cls.test_dir = os.path.realpath(tempfile.mkdtemp())
        cls.gnupg_home = os.path.join(cls.test_dir, "rsa")
        shutil.copytree(gpg_keyring_path, cls.gnupg_home)
        os.chdir(cls.test_dir)

    @classmethod
    def tearDownClass(cls):
        """Change back to initial working dir and remove temp test directory."""

        os.chdir(cls.working_dir)
        shutil.rmtree(cls.test_dir)

    def test_gpg_sign_and_verify_object_with_default_key(self):
        """Create a signature using the default key on the keyring."""
        # pylint: disable=protected-access
        signer = GPGSigner(homedir=self.gnupg_home)
        signature = signer.sign(self.test_data)

        signature_dict = GPGSigner._to_gpg_sig(signature)
        key_data = export_pubkey(self.default_keyid, self.gnupg_home)

        self.assertTrue(verify_sig(signature_dict, key_data, self.test_data))
        self.assertFalse(verify_sig(signature_dict, key_data, self.wrong_data))

    def test_gpg_sign_and_verify_object(self):
        """Create a signature using a specific key on the keyring."""
        # pylint: disable=protected-access
        signer = GPGSigner(self.signing_subkey_keyid, self.gnupg_home)
        signature = signer.sign(self.test_data)

        signature_dict = GPGSigner._to_gpg_sig(signature)
        key_data = export_pubkey(self.signing_subkey_keyid, self.gnupg_home)

        self.assertTrue(verify_sig(signature_dict, key_data, self.test_data))
        self.assertFalse(verify_sig(signature_dict, key_data, self.wrong_data))

    def test_gpg_signature_data_structure(self):
        """Test custom fields and legacy data structure in gpg signatures."""
        # pylint: disable=protected-access
        signer = GPGSigner(homedir=self.gnupg_home)
        sig = signer.sign(self.test_data)
        self.assertIn("other_headers", sig.unrecognized_fields)

        sig_dict = GPGSigner._to_gpg_sig(sig)
        self.assertIn("signature", sig_dict)
        self.assertNotIn("sig", sig_dict)
        sig2 = GPGSigner._from_gpg_sig(sig_dict)
        self.assertEqual(sig, sig2)


# Run the unit tests.
if __name__ == "__main__":
    unittest.main()
